@@ -34,6 +34,8 @@
 #include <sys/errno.h>
 #include <sys/times.h>
 
+#define makecond() allocnode(N_COND)
+
 /* builtin flags */
 
 #define BINF_PLUSOPTS		1		/* +xyz legal */
@@ -126,7 +128,7 @@ struct bincmd builtins[] = {
 	"hash",bin_hash,2,2,0,0,"r",NULL,
 	"history",bin_fc,0,-1,0,BIN_FC,"nr","l",
 	"integer",bin_typeset,0,-1,BINF_TYPEOPTS,0,"LRZlrtux","i",
-	"jobs",bin_fg,0,-1,0,BIN_JOBS,"lp",NULL,
+	"jobs",bin_fg,0,-1,0,BIN_JOBS,"lpZ",NULL,
 	"kill",bin_kill,0,-1,0,0,NULL,NULL,
 	"let",bin_let,1,-1,0,0,NULL,NULL,
 	"limit",bin_limit,0,-1,0,0,"sh",NULL,
@@ -144,7 +146,7 @@ struct bincmd builtins[] = {
 	"rehash",bin_rehash,0,0,0,0,NULL,NULL,
 	"return",bin_break,0,1,0,BIN_RETURN,NULL,NULL,
 	"sched",bin_sched,0,-1,0,0,NULL,NULL,
-	"set",bin_set,0,-1,BINF_SETOPTS|BINF_PLUSOPTS,0,"0123456789ABCDEFGHIJKLMNOPQRSTUWXYZaefghijklmnosuvwxy",NULL,
+	"set",bin_set,0,-1,BINF_SETOPTS|BINF_PLUSOPTS,0,"0123456789ABCDEFGHIJKLMNOPQRSTUWXYZaefghijklnosuvwxy",NULL,
 	"setopt",bin_setopt,0,-1,BINF_PLUSOPTS,0,"0123456789BCDEFGHIJKLMNOPQRSTUWXYZaefghijklmnosuvwxy",NULL,
 	"shift",bin_break,0,1,0,BIN_SHIFT,NULL,NULL,
 	"source",bin_dot,1,-1,0,0,NULL,NULL,
@@ -170,6 +172,23 @@ struct bincmd builtins[] = {
 	NULL,NULL,0,0,0,0,NULL,NULL
 	};
 
+/* print options */
+
+static void prtopt()
+{
+struct option *opp;
+
+	if (isset(KSHOPTIONPRINT)) {
+		printf("Current option settings\n");
+		for (opp = optns; opp->name; opp++)
+			printf("%-20s%s\n", opp->name,
+				(opts[opp->id] == OPT_SET) ? "on" : "off");
+	} else
+		for (opp = optns; opp->name; opp++)
+			if (opts[opp->id] == OPT_SET)
+				puts(opp->name);
+}
+
 /* add builtins to the command hash table */
 
 void addbuiltins() /**/
@@ -242,11 +261,10 @@ int num = -1;
 		}
 	switch (func)
 		{
-		case BIN_BREAK:
-			breaks = (num == -1) ? 1 : num;
-			break;
 		case BIN_CONTINUE:
 			contflag = 1;
+		case BIN_BREAK:
+			breaks = (num == -1) ? 1 : num;
 			break;
 		case BIN_LOGOUT:
 			if (!islogin)
@@ -286,6 +304,7 @@ char *name;char **argv;char *ops;int func;
 {
 int job,lng,firstjob = -1,retval = 0;
 
+	if (ops['Z']) { if (*argv) strcpy(hackzero,*argv); return 0; }
 	lng = (ops['l']) ? 1 : (ops['p']) ? 2 : 0;
 	if ((func == BIN_FG || func == BIN_BG) && !jobbing)
 		{
@@ -374,7 +393,6 @@ int job,lng,firstjob = -1,retval = 0;
 						printf(")\n");
 						}
 					fflush(stdout);
-					settyinfo(&jobtab[job].ttyinfo);
 					attachtty(jobtab[job].gleader);
 					}
 				if (stopped)
@@ -486,11 +504,8 @@ int c;
 
 	if (!ops['@'] && !*args)
 		{
-		if (isun)
-			return 0;
-		for (opp = optns; opp->name; opp++)
-			if (opts[opp->id] == OPT_SET)
-				puts(opp->name);
+		if (!isun)
+			prtopt();
 		return 0;
 		}
 	for (opp = optns; opp->name; opp++)
@@ -710,61 +725,64 @@ char *cnam;
 int bin_cd(nam,argv,ops,func) /**/
 char *nam;char **argv;char *ops;int func;
 {
-char *dest,*new;
-Lknode n;
-int dd,esav = 0,t0;
-char *s,buf[MAXPATHLEN];
-int val,pnew = 0;
+char *dest;
 
 	if (func == BIN_CD && isset(AUTOPUSHD))
 		func = BIN_PUSHD;
+	dest = cd_get_dest(nam,argv,ops,func);
+	if (!dest) return 1;
+	dest = cd_do_chdir(nam,dest);
+	if (!dest) return 1;
+	cd_new_pwd(func,dest);
+	return 0;
+}
+
+char *cd_get_dest(nam,argv,ops,func) /**/
+char *nam;char **argv;char *ops;int func;
+{
+char *dest;
+
 	if (!argv[0])
 		if (func == BIN_CD || (func == BIN_PUSHD && isset(PUSHDTOHOME)
 				|| !full(dirstack)))
 			dest = ztrdup(home);
 		else
 			dest = getnode(dirstack);
-	else if (!argv[1])
-		if (argv[0][1] && argv[0][0] == (isset(PUSHDMINUS) ? '-' : '+'))
-			{
+	else if (!argv[1]) {
+		Lknode n;
+		int dd;
+
+		if (argv[0][1] && argv[0][0] == (isset(PUSHDMINUS) ? '-' : '+')) {
 			dd = atoi(argv[0]+1)-1;
-			if (dd < 0)
-				{
+			if (dd < 0) {
 				zerrnam(nam,"bad directory specification",NULL,0);
-				return 1;
-				}
-			for (n = firstnode(dirstack); n && dd; dd--, incnode(n));
-			if (!n)
-				{
-				zerrnam(nam,"no such entry in dir stack",NULL,0);
-				return 1;
-				}
-			dest = remnode(dirstack,n);
+				return NULL;
 			}
-		else if (argv[0][1] && argv[0][0] == (isset(PUSHDMINUS) ? '+' : '-'))
-			{
+			for (n = firstnode(dirstack); n && dd; dd--, incnode(n));
+			if (!n) {
+				zerrnam(nam,"no such entry in dir stack",NULL,0);
+				return NULL;
+			}
+			dest = remnode(dirstack,n);
+		} else if (argv[0][1] && argv[0][0] == (isset(PUSHDMINUS) ? '+' : '-')) {
 			dd = atoi(argv[0]+1);
 			for (n = lastnode(dirstack); n != (Lknode) dirstack && dd;
 					dd--, n = prevnode(n));
-			if (n == (Lknode) dirstack)
-				{
+			if (n == (Lknode) dirstack) {
 				zerrnam(nam,"no such entry in dir stack",NULL,0);
-				return 1;
-				}
-			dest = remnode(dirstack,n);
+				return NULL;
 			}
-		else
+			dest = remnode(dirstack,n);
+		} else
 			dest = ztrdup(strcmp(argv[0],"-") ? argv[0] : oldpwd);
-	else
-		{
+	} else {
 		char *u;
 		int len1,len2,len3;
 
-		if (!(u = ztrstr(cwd,argv[0])))
-			{
+		if (!(u = ztrstr(cwd,argv[0]))) {
 			zerrnam(nam,"string not in pwd: %s",argv[1],0);
-			return 1;
-			}
+			return NULL;
+		}
 		len1 = strlen(argv[0]);
 		len2 = strlen(argv[1]);
 		len3 = u-cwd;
@@ -772,84 +790,130 @@ int val,pnew = 0;
 		strncpy(dest,cwd,len3);
 		strcpy(dest+len3,argv[1]);
 		strcat(dest,u+len1);
-		}
-	if (*dest == '/')
-		{
-		val = chdir(new = dest);
-		esav = errno;
-		}
-	else if (val = (chdir(dest) != -1))
-		new = dest;
-	else
-		{
-		esav = errno;
-		for (t0 = 0; cdpath[t0]; t0++)
-			{
-			sprintf(buf,"%s/%s",cdpath[t0],dest);
-			if ((val = chdir(new = buf)) != -1)
-				{
-				pnew = 1;
-				break;
-				}
-			if (errno != ENOENT && errno != ENOTDIR)
-				zerrnam(nam,"warning: %e: %s",buf,errno);
-			}
-		}
-	if (val == -1 && errno == ENOENT)
-		{
-		if (isset(CDABLEVARS) &&
-				(s = getsparamval(dest,strlen(dest))) && *s == '/')
-			if (chdir(new = s) != -1)
-				{
-				val = 0;
-				pnew = 1;
-				goto goneto;
-				}
-		zerrnam(nam,"%e: %s",dest,esav);
-		free(dest);
-		return 1;
-		}
-goneto:
-	if (val == -1)
-		{
-		zerrnam(nam,"%e: %s",dest,esav);
-		free(dest);
-		return 1;
-		}
-	else
-		{
-		List l;
+	}
+	return dest;
+}
 
-		oldpwd = cwd;
-		if (isset(CHASELINKS))
-			cwd = findcwd(new);
-		else
-			cwd = fixcwd(new);
+char *cd_do_chdir(nam,dest) /**/
+char *nam; char *dest;
+{
+static char buf[MAXPATHLEN], buf2[MAXPATHLEN];
+struct stat sbuf, tbuf;
+int origerr, val, t0;
+
+	if (*dest == '/') {
+		convertwd(dest,buf);
 		free(dest);
-		if (pnew)
-			{
-			printdir(cwd);
-			putchar('\n');
-			}
-		if (func == BIN_PUSHD)
-			{
-			permalloc();
-			pushnode(dirstack,oldpwd);
-			heapalloc();
-			}
-		if (unset(PUSHDSILENT) && func != BIN_CD && isset(INTERACTIVE))
-			pdstack();
-		if (l = getshfunc("chpwd"))
-			newrunlist(l);
+		val = chdir(buf);
+		if (val == -1) {
+			zerrnam(nam,"%e: %s",dest,errno);
+			free(dest);
+			return NULL;
 		}
-	if (dirstacksize != -1 && countnodes(dirstack) >= dirstacksize)
-		{
+		return buf;
+	}
+	for (t0 = -1;; t0++) {
+		if (t0 == -1) {
+			sprintf(buf2,"%s/%s",cwd,dest);
+			strcpy(buf,dest);
+		} else {
+			if (!cdpath[t0]) break;
+			sprintf(buf2,"%s/%s",cdpath[t0],dest);
+			if (!strncmp(buf2,"../",3))
+				sprintf(buf2,"%s/%s/%s",cwd,cdpath[t0],dest);
+		}
+		convertwd(buf2,buf);
+		val = chdir(dest);
+		if (t0 == -1) origerr = errno;
+		if (val != -1) {
+			if (stat(buf,&sbuf) == -1)
+				if (errno == ENOENT) continue;
+				else break;
+			if (stat(".",&tbuf) == -1) break;
+			if (sbuf.st_dev == tbuf.st_dev && sbuf.st_ino == tbuf.st_ino)
+				break;
+			chdir(buf);
+			break;
+		} else {
+			val = chdir(buf);
+			if (val != -1) break;
+		}
+	}
+	if (t0 > -1 && !cdpath[t0]) {
+		char *s;
+		if (isset(CDABLEVARS) &&
+				(s = getsparam(dest)) && *s == '/' && chdir(s) != -1)
+			strcpy(buf,s);
+		else {
+			zerrnam(nam,"%e: %s",dest,origerr);
+			free(dest);
+			return NULL;
+		}
+	}
+	free(dest);
+	if (t0 > -1) {
+		printdir(buf);
+		putchar('\n');
+	}
+	return buf;
+}
+
+void cd_new_pwd(func,s) /**/
+int func; char *s;
+{
+Param pm;
+List l;
+
+	oldpwd = cwd;
+	if (isset(CHASELINKS))
+		cwd = findcwd(s);
+	else
+		cwd = ztrdup(s);
+	if ((pm = gethnode("PWD", paramtab)) &&
+		 (pm->flags & PMFLAG_x) && pm->env)
+		pm->env = replenv(pm->env,cwd);
+	if ((pm = gethnode("OLDPWD", paramtab)) &&
+		 (pm->flags & PMFLAG_x) && pm->env)
+		pm->env = replenv(pm->env,oldpwd);
+	if (func == BIN_PUSHD) {
+		permalloc();
+		pushnode(dirstack,oldpwd);
+		heapalloc();
+	}
+	if (unset(PUSHDSILENT) && func != BIN_CD && isset(INTERACTIVE))
+		pdstack();
+	if (l = getshfunc("chpwd"))
+		newrunlist(l);
+	if (dirstacksize != -1 && countnodes(dirstack) >= dirstacksize) {
 		if (dirstacksize < 2)
 			dirstacksize = 2;
 		else
 			free(remnode(dirstack,lastnode(dirstack)));
+	}
+}
+
+void convertwd(s,t) /**/
+char *s; char *t;
+{
+char *u,*start;
+
+	*t++ = '/';
+	start = t;
+	for (;;) {
+		while (*s == '/') s++;
+		for (u = s; *u && *u != '/'; u++);
+		if (!strncmp(s,".",u-s)) {
+			;
+		} else if (!strncmp(s,"..",u-s)) {
+			while (t != start && *--t != '/');
+		} else {
+			if (t != start) *t++ = '/';
+			struncpy(&t,s,u-s);
 		}
-	return 0;
+		if (!*u) break;
+		s = u;
+	}
+	*t = '\0';
 }
 
 int bin_rehash(name,argv,ops,func) /**/
@@ -1311,7 +1375,7 @@ List list;
 	hungets(s);
 	free(s);
 	strinbeg();
-	if (!(list = parlist()))
+	if (!(list = parse_list()))
 		{
 		hflush();
 		strinend();
@@ -1358,8 +1422,7 @@ int subbed = 0;
 		while (s3 = (char *) ztrstr(s5,s1))
 			{
 			s4 = alloc(1+(s3-s)+strlen(s2)+strlen(s3+strlen(s1)));
-			strncpy(s4,s,s3-s);
-			s4[s3-s] = '\0';
+			ztrncpy(s4,s,s3-s);
 			strcat(s4,s2);
 			s5 = s4+strlen(s4);
 			strcat(s4,s3+strlen(s1));
@@ -1577,14 +1640,8 @@ char *s;struct alias *a;
 void printparam(s,p) /**/
 char *s;Param p;
 {
-	if (showflag)
-		if (showflag == PMFLAG_SPECIAL)
-			{
-			if (p->flags & showflag)
-				return;
-			}
-		else if (!(p->flags & showflag))
-			return;
+	if (showflag && !(p->flags & showflag))
+		return;
 	if (!showflag)
 		{
 		int fgs = p->flags;
@@ -1645,7 +1702,7 @@ struct asgment *asg;
 	for (; *optstr; optstr++,bit <<= 1)
 		if (ops[*optstr] == 1)
 			on |= bit;
-		else if (ops[*optstr] == -1)
+		else if (ops[*optstr] == 2)
 			off |= bit;
 	roff = off;
 	if (ops['f'])
@@ -1702,6 +1759,10 @@ struct asgment *asg;
 		}
 	else while (asg = getasg(*argv++))
 		{
+		if (asg->value && *asg->value == '~') {
+			*asg->value = Tilde;
+			singsub(&asg->value);
+		}
 		pm = (Param) gethnode(asg->name,paramtab);
 		if (pm)
 			{
@@ -1789,7 +1850,10 @@ int nnl = 0;
 		}
 	if (ops['s'])
 		{
+		permalloc();
 		addnode(histlist,join(args,HISTSPACE));
+		addnode(lithistlist,join(args,' '));
+		heapalloc();
 		curhist++;
 		return 0;
 		}
@@ -1877,7 +1941,6 @@ vptr dat;
 int bin_unhash(name,argv,ops,func) /**/
 char *name;char **argv;char *ops;int func;
 {
-vptr dat;
 Cmdnam chn;
 
 	while (*argv)
@@ -2006,6 +2069,7 @@ int bin_vared(name,args,ops,func) /**/
 char *name;char **args;char *ops;int func;
 {
 char *s,*t;
+struct param *pm;
 
 	if (!(s = getsparam(args[0])))
 		{
@@ -2020,7 +2084,11 @@ char *s,*t;
 		return 1;
 	if (t[strlen(t)-1] == '\n')
 		t[strlen(t)-1] = '\0';
-	setsparam(args[0],t);
+	pm = gethnode(args[0],paramtab);
+	if (pmtype(pm) == PMFLAG_A)
+		setaparam(args[0],spacesplit(t));
+	else
+		setsparam(args[0],t);
 	return 0;
 }
 
@@ -2033,7 +2101,7 @@ Lklist args;Cmdnam cnode;
 {
 struct bincmd *b;
 char ops[128],*arg,*pp,*name,**argv,**oargv,*optstr;
-int t0,flags,sense,argc = 0,retval,op;
+int t0,flags,sense,argc = 0,op;
 Lknode n;
 
 	auxdata = NULL;
@@ -2082,6 +2150,7 @@ Lknode n;
 				}
 			else
 				ops['@'] = 1;
+			op = -1;
 			while (*++arg)
 				if (strchr(b->optstr,op = *arg))
 					ops[*arg] = (sense) ? 1 : 2;
@@ -2098,20 +2167,20 @@ Lknode n;
 				int c;
 				
 				if (!arg)
-					{
-					zerr("option name expected after -o",NULL,0);
-					return 1;
-					}
-				c = optlookup(arg);
-				if (c == -1)
-					{
-					zerr("bad option: %s",arg,0);
-					return 1;
-					}
+					prtopt();
 				else
 					{
-					ops[c] = ops['o'];
-					arg = ugetnode(args);
+					c = optlookup(arg);
+					if (c == -1)
+						{
+						zerr("bad option: %s",arg,0);
+						return 1;
+						}
+					else
+						{
+						ops[c] = ops['o'];
+						arg = ugetnode(args);
+						}
 					}
 				}
 			if ((fset(BINF_PRINTOPTS) && ops['R']) || ops['-'])
@@ -2152,15 +2221,19 @@ Lknode n;
 	argv = oargv;
 	if (errflag)
 		return 1;
-	if (argc < b->minargs || (argc > b->maxargs && b->maxargs != -1))
-		{
+	if (argc < b->minargs || (argc > b->maxargs && b->maxargs != -1)) {
 		zerrnam(name,(argc < b->minargs)
 			? "not enough arguments" : "too many arguments",NULL,0);
-		retval = 1;
-		}
-	else
-		retval = (*(b->handlerfunc))(name,argv,ops,b->funcid);
-	return retval;
+		return 1;
+	}
+	if (isset(XTRACE)) {
+		char **pp = argv;
+		fprintf(stderr,"%s%s",(prompt4) ? prompt4 : "",name);
+		while (*pp) fprintf(stderr," %s",*pp++);
+		fputc('\n',stderr);
+		fflush(stderr);
+	}
+	return (*(b->handlerfunc))(name,argv,ops,b->funcid);
 }
 
 struct asgment *getasg(s) /**/
@@ -2248,9 +2321,9 @@ char *name;char **argv;char *ops;int func;
 struct option *opp;
 char **x;
 
-	if (!ops['@'] && !*argv)
+	if (((ops['+'] && ops['-']) || !ops['-']) && !ops['@'] && !*argv)
 		{
-		showflag = PMFLAG_SPECIAL;
+		showflag = ~0;
 		showflag2 = ops['+'];
 		listhtable(paramtab,(HFunc) printparam);
 		}
@@ -2457,7 +2530,9 @@ long t0;
 		case RLIMIT_DATA: printf("data seg size (kbytes)     "); t0 /= 1024; break;
 		case RLIMIT_STACK: printf("stack size (kbytes)        "); t0 /= 1024; break;
 		case RLIMIT_CORE: printf("core file size (blocks)    "); t0 /= 512; break;
+#ifdef RLIMIT_RSS
 		case RLIMIT_RSS: printf("resident set size (kbytes) "); t0 /= 1024; break;
+#endif
 #ifdef RLIMIT_NOFILE
 		case RLIMIT_NOFILE: printf("file descriptors           "); break;
 #endif
@@ -2488,8 +2563,10 @@ int res,hard;
 		res = RLIMIT_STACK;
 	else if (ops['c'])
 		res = RLIMIT_CORE;
+#ifdef RLIMIT_RSS
 	else if (ops['m'])
 		res = RLIMIT_RSS;
+#endif
 #ifdef RLIMIT_NOFILE
 	else if (ops['n'])
 		res = RLIMIT_NOFILE;
@@ -2523,7 +2600,10 @@ int res,hard;
 		switch(res)
 			{
 			case RLIMIT_FSIZE: case RLIMIT_CORE: t0 *= 512; break;
-			case RLIMIT_DATA: case RLIMIT_STACK: case RLIMIT_RSS:
+			case RLIMIT_DATA: case RLIMIT_STACK:
+#ifdef RLIMIT_RSS
+			case RLIMIT_RSS:
+#endif
 				t0 *= 1024; break;
 			}
 		if (hard)
@@ -2757,16 +2837,17 @@ Cond a,b;
 				}
 			else if (tip[2] != TEST_STR)
 				{
-				parcond2(tsp[0],tsp[1],b = makecond());
+				b = par_cond_double(tsp[0],tsp[1]);
 				tip += 2,tsp += 2;
 				return b;
 				}
 			else
 				{
-				parcond3(tsp[0],tsp[1],tsp[2],b = makecond());
+				b = par_cond_triple(tsp[0],tsp[1],tsp[2]);
 				tip += 3,tsp += 3;
 				return b;
 				}
 		}
+   return NULL;
 }
 
