@@ -32,7 +32,11 @@
 
 #include "zsh.h"
 #include <sys/errno.h>
+#ifdef __hpux
+#include <ndir.h>
+#else
 #include <sys/dir.h>
+#endif
 
 #define execerr() { if (forked) exit(1); \
 	closemnodes(mfds); errflag = 1; return; }
@@ -47,7 +51,7 @@ List l;
 	hungets(s);
 	strinbeg();
 	pushheap();
-	if (!(l = parlist()))
+	if (!(l = parse_list()))
 		{
 		strinend();
 		hflush();
@@ -111,6 +115,8 @@ Cmd cmd;
 
 /* execve after handling $_ and #! */
 
+#define POUNDBANGLIMIT 64
+
 void zexecve(pth,argv,ee,b1,b2) /**/
 char *pth;char **argv;int *ee;char *b1;char *b2;
 {
@@ -133,14 +139,14 @@ char **eep;
 	execve(pth,argv,environ);
 	if ((eno = errno) == ENOEXEC)
 		{
-		char buf[33],*ptr,*ptr2,*argv0;
+		char buf[POUNDBANGLIMIT+1],*ptr,*ptr2,*argv0;
 		int fd,ct,t0;
 
 		if ((fd = open(pth,O_RDONLY)) >= 0)
 			{
 			argv0 = *argv;
 			*argv = pth;
-			ct = read(fd,buf,32);
+			ct = read(fd,buf,POUNDBANGLIMIT);
 			close(fd);
 			if (ct > 0)
 				{
@@ -150,7 +156,7 @@ char **eep;
 						for (t0 = 0; t0 != ct; t0++)
 							if (buf[t0] == '\n')
 								buf[t0] = '\0';
-						buf[32] = '\0';
+						buf[POUNDBANGLIMIT] = '\0';
 						for (ptr = buf+2; *ptr && *ptr == ' '; ptr++);
 						for (ptr2 = ptr; *ptr && *ptr != ' '; ptr++);
 						if (*ptr)
@@ -209,6 +215,10 @@ char *z,*s,buf[MAXCMDLEN],buf2[MAXCMDLEN];
 Cmdnam cn;
 int tl,ee = 0;
 
+	if (!full(args)) {
+		zerr("no command");
+		_exit(1);
+	}
 	cn = (Cmdnam) gethnode(peekfirst(args),cmdnamtab);
 	if (cn && cn->type == DISABLED)
 		cn = NULL;
@@ -366,7 +376,7 @@ Sublist l;int how;int last1;
 int ipipe[2],opipe[2];
 
 	ipipe[0] = ipipe[1] = opipe[0] = opipe[1] = 0;
-	sigblock(sigmask(SIGCHLD));
+	blockchld();
 	if ((thisjob = getfreejob()) == -1)
 		return 1;
 	initjob();
@@ -392,13 +402,13 @@ int ipipe[2],opipe[2];
 	if (how == ASYNC)
 		{
 		spawnjob();
-		sigsetmask(0);
+		unblockchld();
 		return 1;
 		}
 	else
 		{
 		waitjobs();
-		sigsetmask(0);
+		unblockchld();
 		if (l->flags & PFLAG_NOT)
 			lastval = !lastval;
 		return !lastval;
@@ -877,8 +887,9 @@ char *text;
 				addfd(forked,save,mfds,fn->fd1,fil,1);
 				}
 			}
+	/* perform postfork substitutions */
+	/* you put this here so that ls *.sdfoij >& /dev/null works, remember? */
 	postfork(args,!(cmd->flags & CFLAG_NOGLOB));
-		/* perform postfork substitutions */
 	if (errflag)
 		{
 		lastval = 1;
@@ -969,7 +980,8 @@ int bkg;
 		}
 	else if (!jobtab[thisjob].gleader)
 		{
-		setpgrp(0L,jobtab[thisjob].gleader = getpid());
+		jobtab[thisjob].gleader = getpid();
+		setpgrp(0L,jobtab[thisjob].gleader);
 		if (!bkg)
 			attachtty(jobtab[thisjob].gleader);
 		}
@@ -1193,9 +1205,16 @@ int pipes[2];
 	if (*cmd == '<')
 		{
 		int stream;
-		char *fi;
+		char *fi,*s,x;
 
-		fi = strdup(cmd+1);
+		for (cmd++; *cmd == ' '; cmd++);
+		for (s = cmd; *s && *s != ' '; s++)
+			if (*s == '\\') s++;
+			else if (*s == '$') *s = String;
+		x = *s;
+		*s = '\0';
+		fi = strdup(cmd);
+		*s = x;
 		if (*fi == '~')
 			*fi = Tilde;
 		else if (*fi == '=')
@@ -1218,6 +1237,7 @@ int pipes[2];
 		{
 		popheap();
 		close(pipes[1]);
+		/* chldsuspend(); */
 		return readoutput(pipes[0],qt);
 		}
 	subsh = 1;
@@ -1267,7 +1287,7 @@ FILE *fin;
 				ptr = (buf = pp)+cnt;
 				}
 			}
-	if (qt && ptr != buf && ptr[-1] == '\n')
+	if (ptr != buf && ptr[-1] == '\n')
 		ptr[-1] = '\0';
 	else
 		*ptr = '\0';
@@ -1488,7 +1508,7 @@ int cj = thisjob;
 
 char *gettemp() /**/
 {
-	return mktemp(strdup("/tmp/zshXXXXXX"));
+	return mktemp(dyncat(tmpprefix,"XXXXXX"));
 }
 
 /* my getwd; all the other ones I tried confused the SIGCHLD handler */
@@ -1625,7 +1645,10 @@ char *s;
 		cc = (Cmdnam) zalloc(sizeof *cc);
 		cc->type = SHFUNC;
 		cc->flags = 0;
-		cc->u.list = (List) dupstruct(cmd->u.list);
+		if (!cmd->u.list)
+			cc->u.list = NULL;
+		else
+			cc->u.list = (List) dupstruct(cmd->u.list);
 		addhnode(ztrdup(s),cc,cmdnamtab,freecmdnam);
 		if (!strncmp(s,"TRAP",4))
 			{
@@ -1723,7 +1746,7 @@ int fd;
 	for (; *pp; pp++)
 		{
 		sprintf(buf,"%s/%s",*pp,s);
-		if (!access(buf,X_OK) && (fd = open(buf,O_RDONLY)) != -1)
+		if (!access(buf,R_OK) && (fd = open(buf,O_RDONLY)) != -1)
 			{
 			int len = lseek(fd,0,2);
 
