@@ -1,30 +1,33 @@
 struct asgment;
 struct utmp;
 
+#include "builtin.pro"
+#include "cond.pro"
+#include "exec.pro"
 #include "glob.pro"
 #include "hist.pro"
-#include "table.pro"
-#include "subst.pro"
-#include "params.pro"
-#include "builtin.pro"
-#include "loop.pro"
-#include "jobs.pro"
-#include "exec.pro"
 #include "init.pro"
+#include "jobs.pro"
 #include "lex.pro"
-#include "parse.pro"
-#include "utils.pro"
-#include "cond.pro"
+#include "loop.pro"
+#include "math.pro"
 #include "mem.pro"
+#include "params.pro"
+#include "parse.pro"
+#include "subst.pro"
+#include "table.pro"
 #include "text.pro"
+#include "utils.pro"
 #include "watch.pro"
-
-#include "zle_emacs.pro"
-#include "zle_utils.pro"
+#include "zle_hist.pro"
 #include "zle_main.pro"
+#include "zle_misc.pro"
+#include "zle_move.pro"
 #include "zle_refresh.pro"
 #include "zle_tricky.pro"
+#include "zle_utils.pro"
 #include "zle_vi.pro"
+#include "zle_word.pro"
 
 char *mktemp DCLPROTO((char *));
 char *malloc DCLPROTO((int));
