@@ -31,7 +31,16 @@
 */
 
 #include "zsh.h"
+
+#ifdef __hpux
+#include <ndir.h>
+#else
+#ifdef SYSV
+#define direct dirent
+#else
 #include <sys/dir.h>
+#endif
+#endif
 #include <sys/errno.h>
 
 #define exists(X) (access(X,0) == 0)
@@ -107,15 +116,19 @@ char *str = getdata(*np);	/* the pattern */
 				else switch (*s++)
 					{
 					case Hat: case '^': sense = 1-sense; break;
+#ifdef S_IFLNK
 					case '@': func = qualmode; data = S_IFLNK; break;
+#endif
+#ifdef S_IFSOCK
 					case '=': func = qualmode; data = S_IFSOCK; break;
+#endif
 #ifdef S_IFIFO
 					case 'p': func = qualmode; data = S_IFIFO; break;
 #endif
 					case '/': func = qualmode; data = S_IFDIR; break;
 					case '.': func = qualmode; data = S_IFREG; break;
 					case '%': func = qualisdev; break;
-					case Star:  func = qualflags; data = 0100; break;
+					case Star:  func = qualiscom; break;
 					case 'R': func = qualflags; data = 0004; break;
 					case 'W': func = qualflags; data = 0002; break;
 					case 'X': func = qualflags; data = 0001; break;
@@ -134,7 +147,7 @@ char *str = getdata(*np);	/* the pattern */
 					}
 				if (func)
 					{
-					if (qualct == 15)
+					if (qualct == QUALCT-1)
 						{
 						zerr("too many qualifiers",NULL,0);
 						return;
@@ -157,6 +170,7 @@ char *str = getdata(*np);	/* the pattern */
 		qualsense[0] = 0;
 		qualct = 1;
 		}
+	qualfuncs[qualct] = NULL;
 	if (*str == '/')	/* pattern has absolute path */
 		{
 		str++;
@@ -180,18 +194,19 @@ char *str = getdata(*np);	/* the pattern */
 	matchptr = matchbuf = (char **) zalloc((matchsz = 16)*sizeof(char *));
 	matchct = 0;
 	scanner(q);		/* do the globbing */
-	if (!matchct && unset(NULLGLOB))
-		if (unset(NONOMATCH))
-			{
+	if (matchct)
+		badcshglob |= 2;
+	else if (unset(NULLGLOB))
+		if (isset(CSHNULLGLOB)) {
+			badcshglob |= 1;
+		} else if (unset(NONOMATCH)) {
 			zerr("no matches found: %s",ostr,0);
 			free(matchbuf);
 			return;
-			}
-		else
-			{
+		} else {
 			*matchptr++ = strdup(ostr);
 			matchct = 1;
-			}
+		}
 	qsort(&matchbuf[0],matchct,sizeof(char *),notstrcmp);
 	matchptr = matchbuf;
 	while (matchct--)			/* insert matches in the arg list */
@@ -276,7 +291,8 @@ char *str;
 	if (str[0] == '%')
 		return 0;
 	for (; *str; str++)
-		if (!strncmp(str,"..../",5))
+		if (str[0] == Star && str[1] == Star &&
+			 str[2] == Star && str[3] == Star && str[4] == '/')
 			return 1;
 		else if (*str == Pound || *str == Hat || *str == Star ||
 				*str == Bar || *str == Inbrack || *str == Inang ||
@@ -287,8 +303,8 @@ char *str;
 
 /* check to see if str is eligible for brace expansion */
 
-int hasbraces(str,fst) /**/
-char *str;int fst;
+int hasbraces(str) /**/
+char *str;
 {
 int mb,bc,cmct1,cmct2;
 char *lbr = NULL;
@@ -460,8 +476,7 @@ int prev;
 			else if (!*str)
 				exit(10);
 		zz = zalloc(prev+(str-str4)+strlen(str2)+1);
-		strncpy(zz,str3,prev);
-		zz[prev] = '\0';
+		ztrncpy(zz,str3,prev);
 		strncat(zz,str4,str-str4);
 		strcat(zz,str2);
 		insnode(list,node,zz);
@@ -590,12 +605,14 @@ char *t,*lng = NULL,cc,*s = *sp;
 
 /* add a component to pathbuf */
 
-void addpath(s) /**/
+static int addpath(s)
 char *s;
 {
+	if (strlen(s)+pathpos >= MAXPATHLEN) return 0;
 	while (pathbuf[pathpos++] = *s++);
 	pathbuf[pathpos-1] = '/';
 	pathbuf[pathpos] = '\0';
+	return 1;
 }
 
 char *getfullpath(s) /**/
@@ -630,7 +647,8 @@ int closure;
 
 				if (errflag)
 					return;
-				addpath(c->str);
+				if (q->closure && !strcmp(c->str,".")) return;
+				if (!addpath(c->str)) return;
 				if (!closure || exists(pathbuf))
 					scanner((q->closure) ? q : q->next);
 				pathbuf[pathpos = oppos] = '\0';
@@ -682,8 +700,8 @@ int closure;
 							if (type3 != S_IFDIR)
 								continue;
 							}
-						addpath(fn);
-						scanner((q->closure) ? q : q->next); /* scan next level */
+						if (addpath(fn))
+							scanner((q->closure) ? q : q->next); /* scan next level */
 						pathbuf[pathpos = oppos] = '\0';
 						}
 					else
@@ -814,55 +832,45 @@ char *pat = c->str;
 			}
 		if (*pat == Hat)
 			return 1-doesmatch(c->next);
-		if (*pat == Inbrack)
-			if (pat[1] == Hat || pat[1] == '^')
-				{
+		if (*pat == Inbrack) {
+			if (!*pptr) break;
+			if (pat[1] == Hat || pat[1] == '^') {
 				pat[1] = Hat;
 				for (pat += 2; *pat != Outbrack && *pat; pat++)
-					if (*pat == '-' && pat[-1] != Hat && pat[1] != Outbrack)
-						{
+					if (*pat == '-' && pat[-1] != Hat && pat[1] != Outbrack) {
 						if (pat[-1] <= *pptr && pat[1] >= *pptr)
 							break;
-						}
-					else if (*pptr == *pat)
-						break;
-				if (!*pat)
-					{
+					} else if (*pptr == *pat) break;
+				if (!*pat) {
 					zerr("something is very wrong.",NULL,0);
 					return 0;
-					}
+				}
 				if (*pat != Outbrack)
 					break;
 				pat++;
 				pptr++;
 				continue;
-				}
-			else
-				{
+			} else {
 				for (pat++; *pat != Outbrack && *pat; pat++)
-					if (*pat == Inpar)
-						{
+					if (*pat == Inpar) {
 						if (minimatch(&pat,&pptr))
 							break;
-						}
-					else if (*pat == '-' && pat[-1] != Inbrack && pat[1] != Outbrack)
-						{
+					} else if (*pat == '-' && pat[-1] != Inbrack &&
+							pat[1] != Outbrack) {
 						if (pat[-1] <= *pptr && pat[1] >= *pptr)
 							break;
-						}
-					else if (*pptr == *pat)
-						break;
-				if (!pat || !*pat)
-					{
+					} else if (*pptr == *pat) break;
+				if (!pat || !*pat) {
 					zerr("oh dear.  that CAN'T be right.",NULL,0);
 					return 0;
-					}
+				}
 				if (*pat == Outbrack)
 					break;
 				for (pptr++; *pat != Outbrack; pat++);
 				pat++;
 				continue;
-				}
+			}
+		}
 		if (*pat == Inang)
 			{
 			int t1,t2,t3;
@@ -927,14 +935,15 @@ Complist parsecomplist() /**/
 Comp c1;
 Complist p1;
 
-	if (pptr[0] == '.' && pptr[1] == '.' && pptr[2] == '.' && pptr[3] ==
-			'.' && pptr[4] == '/')
+	if (pptr[0] == Star && pptr[1] == Star &&
+		 pptr[2] == Star && pptr[3] == Star &&
+		 pptr[4] == '/')
 		{
 		pptr[0] = Inpar;
 		pptr[1] = Star;
 		pptr[2] = '/';
 		pptr[3] = Outpar;
-		pptr[4] = Pound;	/* "..../" -> "( * /)#" */
+		pptr[4] = Pound;	/* "####/" -> "( * /)#" */
 		}
 	if (*pptr == Inpar)
 		{
@@ -1014,7 +1023,7 @@ char *s = c->str = alloc(MAXPATHLEN*2),*ls = NULL;
 			c1 = (Comp) alloc(sizeof *c1);
 			*(c1->str = strdup("?")) = Quest;
 			c1->closure = 1;
-			c2 = parsecomp();
+			if (!(c2 = parsecomp())) return NULL;
 			c1->next = c2;
 			c->next = c1;
 			return c;
@@ -1046,9 +1055,7 @@ char *s = c->str = alloc(MAXPATHLEN*2),*ls = NULL;
 					dpnd = 2;
 					}
 				}
-			c1 = parsecomp();
-			if (!c1)
-				return NULL;
+			if (!(c1 = parsecomp())) return NULL;
 			tail = c1;
 			endp = pptr;
 			pptr = startp;
@@ -1223,3 +1230,10 @@ struct stat *buf;long mod;
 {
 	return buf->st_mode & mod;
 }
+
+int qualiscom(buf,mod) /**/
+struct stat *buf;long mod;
+{
+	return (buf->st_mode & (S_IFMT|S_IEXEC)) == (S_IFREG|S_IEXEC);
+}
+
