@@ -155,11 +155,13 @@ hatskip:
  
 		for (;;)
 			{
-			if (inblank(c) || c == ':' || c == '^' || c == '$' || c == '*'
-					|| c == '%' || c == '}' || lexstop)
-				break;
-			if (ptr != buf && c == '-')
+			if (inblank(c) || c == ';' || c == ':' || c == '^' || c == '$' ||
+					c == '*' || c == '%' || c == '}' || lexstop)
 				break;
+			if (ptr != buf) {
+				if (c == '-') break;
+				if (idigit(buf[0]) && !idigit(c)) break;
+			}
 			*ptr++ = c;
 			if (c == '#' || c == bangchar)
 				{
@@ -256,7 +258,7 @@ hatskip:
 			switch(c)
 				{
 				case 'p':
-					histdone = 2;
+					histdone = HISTFLAG_DONE|HISTFLAG_NOEXEC;
 					break;
 				case 'h':
 					if (!remtpath(&sline))
@@ -296,7 +298,7 @@ hatskip:
 					char *ptr1,*ptr2;
 				
 					del = hgetch();
-					ptr1 = hdynread(del);
+					ptr1 = hdynread2(del);
 					if (!ptr1)
 						habort();
 					ptr2 = hdynread2(del);
@@ -312,10 +314,7 @@ hatskip:
 					}
 				case '&':
 					if (hsubl && hsubr)
-						{
-						if (subst(&sline,hsubl,hsubr,gbal))
-							habort();
-						}
+						subst(&sline,hsubl,hsubr,gbal);
 					else
 						{
 						herrflush();
@@ -364,7 +363,11 @@ hatskip:
 		alstack[alstackind++] = NULL;
 		}
 	hungets(sline);
-	histdone |= 1;
+	histdone |= HISTFLAG_DONE;
+	if (isset(HISTVERIFY)) {
+		histdone |= HISTFLAG_NOEXEC;
+		zrecall = 1;
+	}
 	goto tailrec;
 }
 
@@ -434,7 +437,7 @@ start:
 			pmpt = putprompt(prompt,&plen);
 			pmpt2 = (rprompt) ? putprompt(rprompt,&foo) : NULL;
 			}
-	if (!(interact && isset(SHINSTDIN) && SHTTY != -1) || unset(USEZLE))
+	if (!(interact && isset(SHINSTDIN) && SHTTY != -1 && isset(USEZLE)))
 		{
 		if (interact && isset(SHINSTDIN))
 			write(2,pmpt,strlen(pmpt));
@@ -637,33 +640,28 @@ void hbegin() /**/
 	stophist = isset(NOBANGHIST);
 	lithist = isset(HISTLIT);
 	hline = hptr = zalloc(hlinesz = 32);
-	if (interact && isset(SHINSTDIN) && !strin)
-		{
+	if (interact && isset(SHINSTDIN) && !strin) {
 		inittty();
 		defev = curhist++;
-		while (curhist-firsthist >= histsiz)
-			{
+		while (curhist-firsthist >= histsiz) {
 			free(getnode(histlist));
 			firsthist++;
-			}
-		while (curhist-firstlithist >= lithistsiz)
-			{
+		}
+		while (curhist-firstlithist >= lithistsiz) {
 			free(getnode(lithistlist));
 			firstlithist++;
-			}
+		}
 		permalloc();
 		addnode(histlist,hline);
 		addnode(lithistlist,ztrdup(""));
 		heapalloc();
-		}
-	else
+	} else
 		histremmed = 1;
 }
 
 void inittty() /**/
 {
 	attachtty(mypgrp);
-	settyinfo(&shttyinfo);
 }
 
 /* say we're done using the history mechanism */
@@ -674,17 +672,15 @@ int flag,save = 1;
 
 	if (!hline)
 		return 1;
-	if (!interact || strin || unset(SHINSTDIN))
-		{
+	if (!interact || strin || unset(SHINSTDIN)) {
 		free(hline);
 		return 1;
-		}
+	}
 	flag = histdone;
 	histdone = 0;
 	if (hptr < hline+2)
 		save = 0;
-	else
-		{
+	else {
 		char *s = getdata(lastnode(lithistlist));
 
 		if (*s)
@@ -701,9 +697,8 @@ int flag,save = 1;
 				!strcmp(hline,getdata(prevnode(lastnode(histlist))))) ||
 				(isset(HISTIGNORESPACE) && spaceflag) )
 			save = 0;
-		}
-	if (flag)
-		{
+	}
+	if (flag & HISTFLAG_DONE) {
 		char *ptr,*p;
 
 		p = ptr = ztrdup(hline);
@@ -713,20 +708,17 @@ int flag,save = 1;
 		fprintf(stderr,"%s\n",ptr);
 		fflush(stderr);
 		free(ptr);
-		}
-	if (!save)
-		{
+	}
+	if (!save) {
 		free(hline);
-		if (!histremmed)
-			{
+		if (!histremmed) {
 			remnode(histlist,lastnode(histlist));
 			free(remnode(lithistlist,lastnode(lithistlist)));
 			curhist--;
-			}
-		flag = 0;
 		}
+	}
 	hline = NULL;
-	return !(flag & 2 || errflag);
+	return !(flag & HISTFLAG_NOEXEC || errflag);
 }
 
 /* remove the current line from the history List */
@@ -853,11 +845,11 @@ char **junkptr;
 {
 char *str = *junkptr,*cut;
  
-	if (cut = strrchr(str,'/'))
-		{
-		*cut = '\0';
+	if (cut = strrchr(str,'/')) {
+		if (str != cut) *cut = '\0';
+		else str[1] = '\0';
 		return 1;
-		}
+	}
 	return 0;
 }
  
@@ -921,34 +913,23 @@ char *str = *junkptr;
 	return 1;
 }
 
-int subst(strptr,in,out,gbal) /**/
+void subst(strptr,in,out,gbal) /**/
 char **strptr;char *in;char *out;int gbal;
 {
 char *str = *strptr,*cut,*sptr;
-int ret = 0,off;
+int off;
 
-considered_harmful:
-	if (cut = (char *) ztrstr(str,in))
-		{
+	while (cut = (char *) ztrstr(str,in)) {
 		*cut = '\0';
 		cut += strlen(in);
 		off = cut-*strptr;
 		*strptr = tricat(*strptr,sptr = convamps(out,in),cut);
-		if (gbal)
-			{
+		if (gbal) {
 			str = (char *) *strptr+off+strlen(sptr);
-			ret = 1;
-			goto considered_harmful;
-			}
-		return 0;
+			continue;
 		}
-	if (!ret)
-		{
-		herrflush();
-		zerr("string not found: %s",in,0);
-		return 1;
-		}
-	return 0;
+		break;
+	}
 }
  
 char *convamps(out,in) /**/
@@ -1169,6 +1150,7 @@ char *fm;int *lenp;
 {
 char *ss,*ttyname DCLPROTO((int)),*bl0;
 static char buf1[256],buf2[256],*buf;
+char buf3[MAXPATHLEN];
 int t0,bracepos = 0;
 struct tm *tm = NULL;
 time_t timet;
@@ -1194,27 +1176,48 @@ time_t timet;
 				{
 				case '~':
 					t0 = finddir(cwd);
-					if (t0 != -1)
-						{
+					if (t0 != -1) {
 						*bp++ = '~';
 						stradd(usernames[t0]);
 						stradd(cwd+strlen(userdirs[t0]));
 						break;
-						}
-					if (!strncmp(cwd,home,t0 = strlen(home)) && t0 > 1)
-						{
+					}
+					if (!strncmp(cwd,home,t0 = strlen(home)) && t0 > 1) {
 						*bp++ = '~';
 						stradd(cwd+t0);
 						break;
-						}
+					}
 				case 'd': case '/': stradd(cwd); break;
 				case 'c': case '.':
-					for (ss = cwd+strlen(cwd); ss > cwd; ss--)
-						if (*ss == '/')
-							{
+					t0 = finddir(cwd);
+					if (t0 != -1) {
+						sprintf(buf3,"~%s%s",usernames[t0],
+							cwd+strlen(userdirs[t0]));
+					} else if (!strncmp(cwd,home,t0 = strlen(home)) && t0 > 1) {
+						sprintf(buf3,"~%s",cwd+t0);
+					} else {
+						strcpy(buf3,cwd);
+					}
+					t0 = 1;
+					if (idigit(fm[1])) { t0 = fm[1]-'0'; fm++; }
+					for (ss = buf3+strlen(buf3); ss > buf3; ss--)
+						if (*ss == '/' && !--t0) {
 							ss++;
 							break;
-							}
+						}
+					if (*ss == '/' && ss[1]) ss++;
+					stradd(ss);
+					break;
+				case 'C':
+					strcpy(buf3,cwd);
+					t0 = 1;
+					if (idigit(fm[1])) { t0 = fm[1]-'0'; fm++; }
+					for (ss = buf3+strlen(buf3); ss > buf3; ss--)
+						if (*ss == '/' && !--t0) {
+							ss++;
+							break;
+						}
+					if (*ss == '/' && ss[1]) ss++;
 					stradd(ss);
 					break;
 				case 'h': case '!':
