@@ -85,10 +85,14 @@ struct rusage ru;
 		}
 	for (;;)
 		{
-#ifdef RLIM_INFINITY
+#ifdef SYSV
+		pid = wait(&statusp);
+#else
+#if defined(RLIM_INFINITY)  & ! defined(__hpux)
 		pid = wait3(&statusp,WNOHANG|WUNTRACED,&ru);
 #else
 		pid = wait3(&statusp,WNOHANG|WUNTRACED,NULL);
+#endif
 #endif
 		if (pid == -1)
 			{
@@ -155,32 +159,25 @@ int notrunning = 1,alldone = 1,val,job = jn-jobtab,somestopped = 0;
 		return;
 	jn->stat |= (alldone) ? STAT_CHANGED|STAT_DONE :
 		STAT_CHANGED|STAT_STOPPED;
-	if (!alldone)
+	if (alldone && job == thisjob)
 		{
-		gettyinfo(&jn->ttyinfo);
-		settyinfo(&shttyinfo);
-		}
-	else if (job == thisjob)
-		{
-		if (!val)
-			{
+		if (!val) {
 			gettyinfo(&shttyinfo);
-			sanetty(&shttyinfo);
+			if (interact && isset(SHINSTDIN) && SHTTY != -1 && isset(USEZLE))
+				sanetty(&shttyinfo);
 #ifdef TIOCSWINSZ
 			if (!(columns = shttyinfo.winsize.ws_col))
 				columns = 80;
 			lines = shttyinfo.winsize.ws_row;
 #endif
-			}
-		else
+		} else
 			settyinfo(&shttyinfo);
 		lastval = val;
 		}
-	if (jn->stat & STAT_STOPPED)
-		{
+	if ((jn->stat & (STAT_DONE|STAT_STOPPED)) == STAT_STOPPED) {
 		prevjob = curjob;
 		curjob = job;
-		}
+	}
 	if ((isset(NOTIFY) || job == thisjob) && jn->stat & STAT_LOCKED)
 		printjob(jn,!!isset(LONGLISTJOBS));
 	if (sigtrapped[SIGCHLD] && job != thisjob)
@@ -411,7 +408,6 @@ void initjob() /**/
 {
 	jobtab[thisjob].cwd = ztrdup(cwd);
 	jobtab[thisjob].stat = STAT_INUSE;
-	jobtab[thisjob].ttyinfo = shttyinfo;
 	jobtab[thisjob].gleader = 0;
 }
 
@@ -460,21 +456,6 @@ Job jn;
 	return 1;
 }
 
-/* wait for a SIGCHLD, wait for the handler to execute, and return */
-
-void chldsuspend() /**/
-{
-#ifdef SV_INTERRUPT
-static struct sigvec vec = { handler,sigmask(SIGCHLD),SV_INTERRUPT };
-
-	sigvec(SIGCHLD,&vec,NULL);
-	sigpause(0);
-	signal(SIGCHLD,handler);
-#else
-	sigpause(0);
-#endif
-}
-
 /* wait for a job to finish */
 
 void waitjob(job) /**/
@@ -536,18 +517,14 @@ int t0;
 
 int getfreejob() /**/
 {
-int mask,t0;
+int t0;
 
-	mask = sigblock(sigmask(SIGCHLD));
 	for (t0 = 1; t0 != MAXJOB; t0++)
-		if (!jobtab[t0].stat)
-			{
-			sigsetmask(mask);
+		if (!jobtab[t0].stat) {
 			jobtab[t0].stat |= STAT_INUSE;
 			return t0;
-			}
-	sigsetmask(mask);
-	zerr("job table full",NULL,0);
+		}
+	zerr("job table full or recursion limit exceeded",NULL,0);
 	return -1;
 }
 
@@ -601,7 +578,7 @@ struct process *pn;
 
 void fixsigs() /**/
 {
-	sigsetmask(0);
+	unblockchld();
 }
 
 void printtime(real,ru,desc) /**/
@@ -623,7 +600,8 @@ char *s;
 				case 'P':
 					if (real)
 						fprintf(stderr,"%d%%",
-							(int) (ru->ru_utime.tv_sec+ru->ru_stime.tv_sec) / real);
+							(int) (100*ru->ru_utime.tv_sec+ru->ru_stime.tv_sec)
+								/ real);
 					break;
 				case 'W': fprintf(stderr,"%ld",ru->ru_nswap); break;
 				case 'X': fprintf(stderr,"%ld",ru->ru_ixrss); break;
@@ -664,6 +642,7 @@ void killrunjobs() /**/
 {
 int t0,killed = 0;
 
+	if (isset(NOHUP)) return;
 	for (t0 = 1; t0 != MAXJOB; t0++)
 		if (t0 != thisjob && jobtab[t0].stat &&
 				!(jobtab[t0].stat & STAT_STOPPED))
