@@ -124,14 +124,17 @@ static struct { char *name; char **data; } x5[] = {
 	"PS3",&prompt3,
 	"PS4",&prompt4,
 	"RPS1",&rprompt,
+	"SPROMPT",&sprompt,
 	"TIMEFMT",&timefmt,
+	"TMPPREFIX",&tmpprefix,
 	"WATCHFMT",&watchfmt,
 	"0",&argzero,
 	NULL,NULL
 	}, *p5 = x5;
 /* colonsplit string params */
 static struct { char *name; } x6[] = {
-	"CDPATH","FIGNORE","FPATH","MAILPATH","WATCH",
+	"CDPATH","FIGNORE","FPATH","MAILPATH","WATCH","HOSTS",
+	"HOSTCMDS","OPTCMDS","BINDCMDS","VARCMDS",
 	NULL
 	}, *p6 = x6;
 /* variable array params */
@@ -141,6 +144,11 @@ static struct { char *name; char ***data; } x7[] = {
 	"fpath",&fpath,
 	"mailpath",&mailpath,
 	"watch",&watch,
+	"hosts",&hosts,
+	"hostcmds",&hostcmds,
+	"optcmds",&optcmds,
+	"bindcmds",&bindcmds,
+	"varcmds",&varcmds,
 	"signals",(char ***) &sigptr,
 	"argv",&pparams,
 	"*",&pparams,
@@ -270,6 +278,7 @@ struct param *createparam(name,value,flags) /**/
 char *name;vptr value;int flags;
 {
 struct param *pm;
+char buf[20];
 
 	pm = zcalloc(sizeof *pm);
 	if (isset(ALLEXPORT))
@@ -292,31 +301,34 @@ struct param *pm;
 		pm->u.val = (value) ? matheval(value) : 0;
 		pm->sets.ifn = intsetfn;
 		pm->gets.ifn = intgetfn;
+		sprintf(buf,"%ld",pm->u.val);
+		value = buf;
 		}
 	if (flags & PMFLAG_x)
 		pm->env = addenv(name,value);
-	addhnode(ztrdup(name),pm,paramtab,NULL);
+	addhnode(ztrdup(name),pm,paramtab,freepm);
 	return pm;
 }
 
 int isident(s) /**/
 char *s;
 {
+char *ss;
+
+	for (ss = s; *ss; ss++) if (!iident(*ss)) break;
+	if (!*ss || *ss == '[') return 1;
 	if (*s == Quest)
 		*s = '?';
 	else if (*s == Pound)
 		*s = '#';
-	else if (*s == String)
-		*s = '$';
-	else if (*s == Qstring)
+	else if (*s == String || *s == Qstring)
 		*s = '$';
 	else if (*s == Star)
 		*s = '*';
-	else if (!idigit(*s) && !iident(*s) && *s != '#' && *s != '-' &&
-			*s != '?' && *s != '$' && *s != '_' && *s != '!' &&
-			*s != '@' && *s != '*')
-		return 0;
-	return 1;
+	if (*s == '#' || *s == '-' || *s == '?' || *s == '$' || *s == '_' ||
+		 *s == '!' || *s == '@' || *s == '*')
+		return 1;
+	return 0;
 }
 
 Value getvalue(pptr,bracks) /**/
@@ -326,37 +338,24 @@ char *s = *pptr,*t = *pptr;
 char sav;
 Value v;
 
-	if (idigit(*s))
-		while (idigit(*s)) s++;
-	else if (iident(*s))
-		while (iident(*s)) s++;
-	else if (*s == Quest)
-		*s++ = '?';
-	else if (*s == Pound)
-		*s++ = '#';
-	else if (*s == String)
-		*s++ = '$';
-	else if (*s == Qstring)
-		*s++ = '$';
-	else if (*s == Star)
-		*s++ = '*';
+	if (idigit(*s)) while (idigit(*s)) s++;
+	else if (iident(*s)) while (iident(*s)) s++;
+	else if (*s == Quest) *s++ = '?';
+	else if (*s == Pound) *s++ = '#';
+	else if (*s == String) *s++ = '$';
+	else if (*s == Qstring) *s++ = '$';
+	else if (*s == Star) *s++ = '*';
 	else if (*s == '#' || *s == '-' || *s == '?' || *s == '$' ||
-			*s == '_' || *s == '!' || *s == '@' || *s == '*')
-		s++;
-	else
-		return NULL;
-	if (sav = *s)
-		*s = '\0';
-	if (idigit(*t) && *t != '0')
-		{
+				*s == '_' || *s == '!' || *s == '@' || *s == '*') s++;
+	else return NULL;
+	if (sav = *s) *s = '\0';
+	if (idigit(*t) && *t != '0') {
 		v = (Value) alloc(sizeof *v);
 		v->pm = argvparam;
 		v->a = v->b = atoi(t)-1;
 		if (sav)
 			*s = sav;
-		}
-	else
-		{
+	} else {
 		struct param *pm;
 		int isvarat = !strcmp(t, "@");
 
@@ -370,9 +369,8 @@ Value v;
 		if (pmtype(pm) == PMFLAG_A)
 			v->isarr = isvarat ? -1 : 1;
 		v->pm = pm;
-		v->a = v->b = -1;
-		if (bracks && (*s == '[' || *s == Inbrack))
-			{
+		v->a = 0; v->b = -1;
+		if (bracks && (*s == '[' || *s == Inbrack)) {
 			int a,b;
 			char *olds = s,*t;
 
@@ -382,40 +380,31 @@ Value v;
 					*t = ztokens[*t-Pound];
 			if (*t == Outbrack)
 				*t = ']';
-			if ((s[0] == '*' || s[0] == '@')  && s[1] == ']')
-				{
+			if ((s[0] == '*' || s[0] == '@')  && s[1] == ']') {
 				v->isarr = (s[0] == '*') ? 1 : -1;
 				v->a = 0;
 				v->b = -1;
 				s += 2;
-				}
-			else
-				{
-				a = mathevalarg(s,&s)-1;
-				if (a < 0)
-					a = -1;
-				else if (*s == ',' || *s == Comma)
-					{
+			} else {
+				a = mathevalarg(s,&s);
+				if (a > 0) a--;
+				if (*s == ',' || *s == Comma) {
 					s++;
-					b = mathevalarg(s,&s)-1;
-					if (b < 0 || b < a)
-						b = -1;
-					}
-				else
+					b = mathevalarg(s,&s);
+					if (b > 0) b--;
+				} else
 					b = a;
-				if (*s == ']')
-					{
+				if (*s == ']') {
 					s++;
-					if (v->isarr && a == b && a != -1)
+					if (v->isarr && a == b)
 						v->isarr = 0;
 					v->a = a;
 					v->b = b;
-					}
-				else
+				} else
 					s = olds;
-				}
 			}
 		}
+	}
 	if (!bracks && *s)
 		return NULL;
 	*pptr = s;
@@ -425,32 +414,29 @@ Value v;
 char *getstrvalue(v) /**/
 Value v;
 {
-char *s;
+char *s,**ss;
 static char buf[20];
 
 	if (!v)
 		return "";
-	if (pmtype(v->pm) != PMFLAG_A)
-		{
+	if (pmtype(v->pm) != PMFLAG_A) {
 		if ((pmtype(v->pm) == PMFLAG_i))
 			convbase(s = buf,v->pm->gets.ifn(v->pm),v->pm->ct);
 		else
 			s = v->pm->gets.cfn(v->pm);
-		if (v->a != -1)
-			{
-			s = (v->a > strlen(s)) ? strdup("") : strdup(s+v->a);
-			if (v->b != -1 && v->b-v->a < strlen(s))
-				s[v->b-v->a+1] = '\0';
-			}
-		}
-	else if (v->isarr)
-		s = spacejoin(v->pm->gets.afn(v->pm));
-	else
-		{
-		char **ss = v->pm->gets.afn(v->pm);
+		if (v->a == 0 && v->b == -1) return s;
+		if (v->a < 0) v->a += strlen(s);
+		if (v->b < 0) v->b += strlen(s);
+		s = (v->a > strlen(s)) ? strdup("") : strdup(s+v->a);
+		if (v->b < v->a) s[0] = '\0';
+		else if (v->b-v->a < strlen(s)) s[v->b-v->a+1] = '\0';
+		return s;
+	}
+	if (v->isarr) return spacejoin(v->pm->gets.afn(v->pm));
 
-		s = (v->a >= arrlen(ss)) ? "" : ss[v->a];
-		}
+	ss = v->pm->gets.afn(v->pm);
+	if (v->a < 0) v->a += arrlen(ss);
+	s = (v->a >= arrlen(ss) || v->a < 0) ? "" : ss[v->a];
 	return s;
 }
 
@@ -463,30 +449,34 @@ static char *nular[] = { "", NULL };
 	if (!v)
 		return arrdup(nular);
 	s = v->pm->gets.afn(v->pm);
-	if (v->a != -1)
-		{
-		if (v->a > arrlen(s))
-			s = arrdup(nular);
-		else
-			s = arrdup(s)+v->a;
-		if (v->b != -1 && v->b-v->a < arrlen(s))
-			s[v->b-v->a+1] = NULL;
-		}
+	if (v->a == 0 && v->b == -1) return s;
+	if (v->a < 0) v->a += arrlen(s);
+	if (v->b < 0) v->b += arrlen(s);
+	if (v->a > arrlen(s) || v->a < 0)
+		s = arrdup(nular);
+	else
+		s = arrdup(s)+v->a;
+	if (v->b < v->a) s[0] = NULL;
+	else if (v->b-v->a < arrlen(s)) s[v->b-v->a+1] = NULL;
 	return s;
 }
 
 long getintvalue(v) /**/
 Value v;
 {
+char **ss;
+
 	if (!v || v->isarr)
 		return 0;
-	if (pmtype(v->pm) != PMFLAG_A)
-		{
+	if (pmtype(v->pm) != PMFLAG_A) {
 		if (pmtype(v->pm) == PMFLAG_i)
 			return v->pm->gets.ifn(v->pm);
 		return atol(v->pm->gets.cfn(v->pm));
-		}
-	return atol(v->pm->gets.afn(v->pm)[v->a]);
+	}
+	ss = v->pm->gets.afn(v->pm);
+	if (v->a < 0) v->a += arrlen(ss);
+	if (v->a < 0 || v->a > arrlen(ss)) return 0;
+	return atol(ss[v->a]);
 }
 
 void setstrvalue(v,val) /**/
@@ -498,24 +488,29 @@ char *s;
 		return;
 	if ((s = v->pm->env) && val)
 		v->pm->env = replenv(v->pm->env,val);
-	switch (pmtype(v->pm))
-		{
+	switch (pmtype(v->pm)) {
 		case PMFLAG_s:
-			if (v->a == -1)
+			if (v->a == 0 && v->b == -1)
 				(v->pm->sets.cfn)(v->pm,val);
-			else
-				{
+			else {
 				char *z,*y,*x;
 
 				z = strdup((v->pm->gets.cfn)(v->pm));
+				if (v->a < 0) {
+					v->a += strlen(z);
+					if (v->a < 0) v->a = 0;
+				}
+				if (v->a > strlen(z)) v->a = strlen(z);
+				if (v->b < 0) v->b += strlen(z);
+				if (v->b <= v->a) v->b = v->a-1;
 				z[v->a] = '\0';
-				y = (v->b == -1) ? "" : z+v->b+1;
+				y = z+v->b+1;
 				x = zalloc(strlen(z)+strlen(y)+strlen(val)+1);
 				strcpy(x,z);
 				strcat(x,val);
 				strcat(x,y);
 				(v->pm->sets.cfn)(v->pm,z);
-				}
+			}
 			if (v->pm->flags & (PMFLAG_L|PMFLAG_R|PMFLAG_Z) && !v->pm->ct)
 				v->pm->ct = strlen(val);
 			break;
@@ -526,15 +521,18 @@ char *s;
 			free(val);
 			break;
 		case PMFLAG_A:
-			if (v->a == -1 || v->a != v->b)
+			if (v->a != v->b)
 				zerr("illegal array assignment",NULL,0);
-			else
-				{
+			else {
 				char **ss = (v->pm->gets.afn)(v->pm);
 				int ac,ad,t0;
 
-				if (v->a >= (ac = arrlen(ss)))
-					{
+				ac = arrlen(ss);
+				if (v->a < 0) {
+					v->a += ac;
+					if (v->a < 0) v->a = 0;
+				}
+				if (v->a >= ac) {
 					char **st = ss;
 
 					ad = v->a+1;
@@ -545,13 +543,13 @@ char *s;
 					while (ac < ad)
 						ss[ac++] = ztrdup("");
 					ss[ac] = NULL;
-					}
+				}
 				if (ss[v->a]) free(ss[v->a]);
 				ss[v->a] = val;
 				(v->pm->sets.afn)(v->pm,ss);
-				}
+			}
 			break;
-		}
+	}
 }
 
 void setintvalue(v,val) /**/
@@ -561,6 +559,10 @@ char buf[20];
 
 	if (v->pm->flags & PMFLAG_r)
 		return;
+	if (v->pm->env) {
+		sprintf(buf,"%ld",val);
+		v->pm->env = replenv(v->pm->env,buf);
+	}
 	switch (pmtype(v->pm))
 		{
 		case PMFLAG_s:
@@ -639,6 +641,11 @@ char *t = s;
 		}
 	if (!(v = getvalue(&s,1)) || *s)
 		return createparam(t,val,PMFLAG_s);
+	if ((v->pm->flags & PMTYPE) != PMFLAG_s &&
+			!(v->pm->flags & PMFLAG_SPECIAL)) {
+		unsetparam(s);
+		return createparam(t,val,PMFLAG_s);
+	}
 	setstrvalue(v,val);
 	return v->pm;
 }
@@ -656,6 +663,11 @@ char *t = s;
 		}
 	if (!(v = getvalue(&s,1)) || *s)
 		return createparam(t,val,PMFLAG_A);
+	if ((v->pm->flags & PMTYPE) != PMFLAG_A &&
+			!(v->pm->flags & PMFLAG_SPECIAL)) {
+		unsetparam(s);
+		return createparam(t,val,PMFLAG_A);
+	}
 	setarrvalue(v,val);
 	return v->pm;
 }
@@ -707,7 +719,7 @@ Param pm;
 	if (pmtype(pm) == PMFLAG_s && (pm->flags & PMFLAG_x))
 		delenv(pm->env);
 	if (!(pm->flags & PMFLAG_SPECIAL))
-		remhnode(s,paramtab); /* storage leak - no freepm */
+		freepm(remhnode(s,paramtab));
 	unsetflag = 0;
 }
 
@@ -788,7 +800,12 @@ Param pm;char *x;
 char *strvargetfn(pm) /**/
 Param pm;
 {
-	return *((char **) pm->data);
+char *s;
+
+	s = *((char **) pm->data);
+	if (!s)
+		return "";
+	return s;
 }
 
 char *strconstgetfn(pm) /**/
@@ -1059,7 +1076,7 @@ Param pm;
 void homesetfn(pm,x) /**/
 Param pm;char *x;
 {
-	if (home = xsymlink(x))
+	if (isset(CHASELINKS) && (home = xsymlink(x)))
 		free(x);
 	else
 		home = x;
@@ -1184,28 +1201,35 @@ int ct;
 	*envp2 = NULL;
 	envp = environ;
 	environ = envp2 = envp3;
-	for (; *envp; envp++,envp2++)
-		{
+	for (; *envp; envp++,envp2++) {
 		for (str = *envp; *str && *str != '='; str++);
-		if (*str == '=')
-			{
+		if (*str == '=') {
+			char *iname;
+
 			*str = '\0';
 			if (isident(*envp))
-				pm = setsparam(ztrdup(*envp),ztrdup(str+1));
+				pm = setsparam(iname = ztrdup(*envp),ztrdup(str+1));
 			*str = '=';
-			if (pm)
-				{
+			if (pm) {
 				pm->flags |= PMFLAG_x;
 				pm->env = *envp2;
-				}
+				if (pm->flags & PMFLAG_SPECIAL)
+					pm->env = replenv(pm->env,getsparam(iname));
 			}
 		}
+	}
 	pm = gethnode("HOME",paramtab);
 	if (!(pm->flags & PMFLAG_x))
 		{
 		pm->flags |= PMFLAG_x;
 		pm->env = addenv("HOME",home);
 		}
+	pm = gethnode("PWD",paramtab);
+	if (!(pm->flags & PMFLAG_x))
+		{
+		pm->flags |= PMFLAG_x;
+		pm->env = addenv("PWD",cwd);
+		}
 	pm = gethnode("SHLVL",paramtab);
 	if (!(pm->flags & PMFLAG_x))
 		pm->flags |= PMFLAG_x;
