@@ -32,9 +32,25 @@
 #include "zsh.h"
 #include <pwd.h>
 #include <errno.h>
+#ifdef __hpux
+#include <ndir.h>
+#else
+#ifndef SYSV
 #include <sys/dir.h>
+#endif
+#endif
 #include <fcntl.h>
 
+#ifdef SYSV
+#define direct dirent
+#undef TIOCGWINSZ
+int readlink(s,t,z)
+char *s;char *t;int z;
+{
+return -1;
+}
+#endif
+
 /* source a file */
 
 int source(s) /**/
@@ -140,7 +156,7 @@ char *cmd; char *fmt; char *str;int num;
 			}
 		else
 			putc(*fmt++,stderr);
-	if (unset(SHINSTDIN))
+	if (unset(SHINSTDIN) && lineno)
 		fprintf(stderr," [%ld]\n",lineno);
 	else
 		putc('\n',stderr);
@@ -178,7 +194,7 @@ int c;FILE *f;
 	else
 		{
 		putc('^',f);
-		putc(c|'A',f);
+		putc(c|'@',f);
 		}
 }
 
@@ -191,7 +207,6 @@ static struct sigvec vec = { handler,sigmask(SIGINT),SV_INTERRUPT };
 
 	if (interact)
 		sigvec(SIGINT,&vec,NULL);
-	sigsetmask(0);
 #else
 	if (interact)
 		signal(SIGINT,handler);
@@ -208,14 +223,9 @@ void holdintr() /**/
 #ifdef SV_INTERRUPT
 static struct sigvec vec = { handler,sigmask(SIGINT),0 };
 
-	if (interact)
-		{
-		sigvec(SIGINT,&vec,NULL);
-		sigsetmask(0);
-		}
+	if (interact) sigvec(SIGINT,&vec,NULL);
 #else
-	if (interact)
-		signal(SIGINT,SIG_IGN);
+	if (interact) signal(SIGINT,SIG_IGN);
 #endif
 }
 
@@ -246,6 +256,7 @@ char *t;
 
 static char xbuf[MAXPATHLEN];
 
+/* get rid of this */
 char *fixcwd(s) /**/
 char *s;
 {
@@ -257,7 +268,7 @@ char *t;
 		t = ztrdup(s);
 	else
 		t = tricat((cwd[1]) ? cwd : "","/",s);
-	(void) xsymlinks(t+1); 
+	(void) xsymlinks(t+1,0); 
 	free(t);
 	if (!*xbuf)
 		strcpy(xbuf,"/");
@@ -288,7 +299,7 @@ char *s;
 	if (*s != '/')
 		return NULL;
 	strcpy(xbuf,"");
-	if (xsymlinks(s+1))
+	if (xsymlinks(s+1,1))
 		return ztrdup(s);
 	if (!*xbuf)
 		return ztrdup("/");
@@ -332,8 +343,11 @@ char *s;
 	return readlink(s,NULL,0) == 0;
 }
 
-int xsymlinks(s) /**/
-char *s;
+/* expands symlinks and .. or . expressions */
+/* if flag = 0, only expand .. and . expressions */
+
+int xsymlinks(s,flag) /**/
+char *s;int flag;
 {
 char **pp,**opp;
 char xbuf2[MAXPATHLEN],xbuf3[MAXPATHLEN];
@@ -368,7 +382,7 @@ int t0;
 			}
 		sprintf(xbuf2,"%s/%s",xbuf,*pp);
 		t0 = readlink(xbuf2,xbuf3,MAXPATHLEN);
-		if (t0 == -1)
+		if (t0 == -1 || !flag)
 			{
 			strcat(xbuf,"/");
 			strcat(xbuf,*pp);
@@ -380,11 +394,11 @@ int t0;
 			if (*xbuf3 == '/')
 				{
 				strcpy(xbuf,"");
-				if (xsymlinks(xbuf3+1))
+				if (xsymlinks(xbuf3+1,flag))
 					return 1;
 				}
 			else
-				if (xsymlinks(xbuf3))
+				if (xsymlinks(xbuf3,flag))
 					return 1;
 			free(*pp);
 			}
@@ -704,17 +718,15 @@ Hashtab oldcnt;
 	oldcnt = cmdnamtab;
 	permalloc();
 	cmdnamtab = newhtable(101);
-	if (pathsuppress)
-		{
+	if (pathsuppress) {
 		addbuiltins();
-		if (oldcnt)
-			{
+		if (oldcnt) {
 			listhtable(oldcnt,(HFunc) saveoldfuncs);
 			freehtab(oldcnt,freecmdnam);
-			}
+		}
 		lastalloc();
 		return;
-		}
+	}
 	holdintr();
 	for (t0 = 0; path[t0]; t0++)
 		if (!strcmp(".",path[t0]))
@@ -755,12 +767,10 @@ vptr a;
 {
 struct cmdnam *c = (struct cmdnam *) a;
 
-	if (c->type == SHFUNC)
-		{
+	if (c->type == SHFUNC) {
 		if (c->u.list)
 			freestruct(c->u.list);
-		}
-	else if (c->type != BUILTIN)
+	} else if (c->type != BUILTIN && c->type != DISABLED)
 		free(c->u.nam);
 	free(c);
 }
@@ -799,7 +809,11 @@ struct ttyinfo *ti;
 	if (SHTTY != -1)
 		{
 #ifdef TERMIOS
+#ifdef __hpux
+		if (tcgetattr(SHTTY,&ti->termios) == -1)
+#else
 		if (ioctl(SHTTY,TCGETS,&ti->termios) == -1)
+#endif
 			zerr("bad tcgets: %e",NULL,errno);
 #else
 #ifdef TERMIO
@@ -824,7 +838,11 @@ struct ttyinfo *ti;
 	if (SHTTY != -1)
 		{
 #ifdef TERMIOS
+#ifdef __hpux
+		if (tcsetattr(SHTTY, TCSADRAIN, &ti->termios) == -1)
+#else
 		if (ioctl(SHTTY,TCSETS,&ti->termios) == -1)
+#endif
 			zerr("settyinfo: %e",NULL,errno);
 #else
 #ifdef TERMIO
@@ -1048,6 +1066,15 @@ int sav;
 	sigtrapped[sig] = sav;
 }
 
+/* copy len chars from t into s, and null terminate */
+
+void ztrncpy(s,t,len) /**/
+char *s;char *t;int len;
+{
+	while (len--) *s++ = *t++;
+	*s = '\0';
+}
+
 /* copy t into *s and update s */
 
 void strucpy(s,t) /**/
@@ -1073,7 +1100,7 @@ char *u = *s;
 void checkrmall() /**/
 {
 	fflush(stdin);
-	fprintf(stderr,"zsh: are you sure you want to delete all the files? ");
+	fprintf(stderr,"zsh: sure you want to delete all the files in %s? ",cwd);
 	fflush(stderr);
 	feep();
 	errflag |= (getquery() != 'y');
@@ -1082,18 +1109,14 @@ void checkrmall() /**/
 int getquery() /**/
 {
 char c;
-int yes = 'q';
 
 	setcbreak();
 	if (read(SHTTY,&c,1) == 1)
-		if (c == 'y' || c == 'Y' || c == '\t')
-			yes = 'y';
-		else if (c == 'n')
-			yes = 'n';
+		if (c == 'y' || c == 'Y' || c == '\t') c = 'y';
 	unsetcbreak();
 	if (c != '\n')
 		write(2,"\n",1);
-	return yes;
+	return (int) c;
 }
 
 static int d;
@@ -1120,9 +1143,10 @@ int nd;
 }
 
 /* spellcheck a word */
+/* fix s and s2 ; if s2 is non-null, fix the history list too */
 
-void spckword(s,s2,cmd,ask) /**/
-char **s;char **s2;int cmd;int ask;
+void spckword(s,s2,tptr,cmd,ask) /**/
+char **s;char **s2;char **tptr;int cmd;int ask;
 {
 char *t,*u;
 int x;
@@ -1144,30 +1168,48 @@ int x;
 			break;
 	if ((u = spname(*s)) != *s)
 		best = u;
-	else if (!*t && !cmd)
-		{
+	else if (!*t && !cmd) {
 		guess = *s;
 		d = 100;
 		listhtable(aliastab,spscan);
 		listhtable(cmdnamtab,spscan);
-		}
-	if (best && strlen(best) > 1)
-		{
+	}
+	if (best && strlen(best) > 1 && strcmp(best,*s)) {
 		if (ask) {
-			fprintf(stderr,"zsh: correct to `%s' (y/n)? ",best);
+			char *p = sprompt;
+			while (*p) {
+				if (*p == '%') {
+					if (p[1] == 'r') fprintf(stderr,"%s",best);
+					if (p[1] == 's') fprintf(stderr,"%s",*s);
+					if (p[1]) p++;
+				} else fputc(*p,stderr);
+				p++;
+			}
 			fflush(stderr);
 			feep();
 			x = getquery();
-		}
-		else
+		} else
 			x = 'y';
-		if (x == 'y')
-			{
+		if (x == 'y') {
 			*s = strdup(best);
-			if (s2)
+			if (s2) {
+				if (*tptr && !strcmp(hlastw,*s2)) {
+					char *z;
+					hptr = hlastw;
+					for (z = best; *z; z++) hwaddc(*z);
+					hwaddc(HISTSPACE);
+					*tptr = hptr-1;
+					**tptr = '\0';
+				}
 				*s2 = strdup(best);
 			}
+		} else if (x == 'a') {
+			histdone |= HISTFLAG_NOEXEC;
+		} else if (x == 'e') {
+			histdone |= HISTFLAG_NOEXEC;
+			zrecall = 1;
 		}
+	}
 }
 
 int ztrftime(buf,bufsize,fmt,tm) /**/
@@ -1209,8 +1251,8 @@ static char *lstr[] = {"12"," 1"," 2"," 3"," 4"," 5"," 6"," 7"," 8"," 9",
 					strucpy(&buf,lstr[tm->tm_hour%12]);
 					break;
 				case 'm':
-					*buf++ = '0'+tm->tm_mon/10;
-					*buf++ = '0'+tm->tm_mon%10;
+					*buf++ = '0'+(tm->tm_mon+1)/10;
+					*buf++ = '0'+(tm->tm_mon+1)%10;
 					break;
 				case 'M':
 					*buf++ = '0'+tm->tm_min/10;
@@ -1285,8 +1327,7 @@ char *t,**ret,**ptr;
 		{
 		for (s = t; *t && *t != ':'; t++);
 		*ptr = zalloc((t-s)+1);
-		strncpy(*ptr,s,(t-s)+1);
-		(*ptr++)[t-s] = '\0';
+		ztrncpy(*ptr++,s,t-s);
 		}
 	while (*t++);
 	*ptr = NULL;
@@ -1308,8 +1349,7 @@ char *t,**ret,**ptr;
 		{
 		for (s = t; *t && !inblank(*t); t++);
 		*ptr = zalloc((t-s)+1);
-		strncpy(*ptr,s,(t-s)+1);
-		(*ptr++)[t-s] = '\0';
+		ztrncpy(*ptr++,s,t-s);
 		}
 	while (*t++);
 	*ptr = NULL;
@@ -1521,7 +1561,7 @@ char *spname (oldname) /**/
 char *oldname;
 {
 	char *p,guess[MAXPATHLEN+1],best[MAXPATHLEN+1];
-	char newname[MAXPATHLEN+1];
+	static char newname[MAXPATHLEN+1];
 	char *new = newname, *old = oldname;
 
 	for (;;)
@@ -1657,3 +1697,20 @@ int c;
 	return (c >= 'a' && c <= 'z') ? c-'a'+'A' : c;
 }
 
+#ifdef SYSV
+#include <sys/utsname.h>
+
+int gethostname(nameptr, maxlength)
+char *nameptr;
+int maxlength;
+{
+struct utsname *name;
+int result;
+
+	result = uname(name);
+	if (result >= 0) {
+		strcpy(nameptr,name->sysname);
+		return 0;
+	} else return -1;
+}
+#endif
