@@ -37,124 +37,148 @@
 struct zlecmd zlecmds[] = {
 "accept-and-hold",acceptandhold,0,
 "accept-and-infer-next-history",acceptandinfernexthistory,0,
-"accept-and-menu-complete", acceptandmenucomplete, ZLE_MOD|ZLE_MENUCMP,
+"accept-and-menu-complete", acceptandmenucomplete, ZLE_MENUCMP,
 "accept-line",acceptline,0,
-"accept-line-and-down-history",acceptlineanddownhistory,ZLE_MOD,
-"backward-char",backwardchar,ZLE_MOVE,
-"backward-delete-char",backwarddeletechar,ZLE_INSMOD,
-"backward-delete-word",backwarddeleteword,ZLE_MOD,
-"backward-kill-line",backwardkillline,ZLE_MOD|ZLE_KILL,
-"backward-kill-word",backwardkillword,ZLE_MOD|ZLE_KILL,
-"backward-word",backwardword,ZLE_MOVE,
-"beginning-of-buffer-or-history",beginningofbufferorhistory,ZLE_LINEMOVE,
-"beginning-of-history",beginningofhistory,ZLE_MOD,
-"beginning-of-line",beginningofline,ZLE_MOVE,
-"beginning-of-line-hist",beginningoflinehist,ZLE_LINEMOVE,
-"capitalize-word",capitalizeword,ZLE_MOD,
+"accept-line-and-down-history",acceptlineanddownhistory,0,
+"backward-char",backwardchar,ZLE_MOVEMENT,
+"backward-delete-char",backwarddeletechar,0,
+"backward-delete-word",backwarddeleteword,0,
+"backward-kill-line",backwardkillline,ZLE_KILL,
+"backward-kill-word",backwardkillword,ZLE_KILL,
+"backward-word",backwardword,ZLE_MOVEMENT,
+"beginning-of-buffer-or-history",beginningofbufferorhistory,ZLE_MOVEMENT,
+"beginning-of-history",beginningofhistory,0,
+"beginning-of-line",beginningofline,ZLE_MOVEMENT,
+"beginning-of-line-hist",beginningoflinehist,ZLE_MOVEMENT,
+"capitalize-word",capitalizeword,0,
 "clear-screen",clearscreen,0,
-"complete-word",completeword,ZLE_MOD|ZLE_MENUCMP,
-"copy-prev-word",copyprevword,ZLE_MOD,
+"complete-word",completeword,ZLE_MENUCMP,
+"copy-prev-word",copyprevword,0,
 "copy-region-as-kill",copyregionaskill,ZLE_KILL,
-"delete-char",deletechar,ZLE_INSMOD,
-"delete-char-or-list",deletecharorlist,ZLE_INSMOD|ZLE_MENUCMP,
-"delete-word",deleteword,ZLE_MOD,
+"delete-char",deletechar,0,
+"delete-char-or-list",deletecharorlist,ZLE_MENUCMP,
+"delete-word",deleteword,0,
 "digit-argument",digitargument,ZLE_ARG,
-"down-case-word",downcaseword,ZLE_MOD,
-"down-history",downhistory,ZLE_MOD,
-"down-line-or-history",downlineorhistory,ZLE_LINEMOVE2,
-"end-of-buffer-or-history",endofbufferorhistory,ZLE_LINEMOVE,
-"end-of-history",endofhistory,ZLE_MOD,
-"end-of-line",endofline,ZLE_MOVE,
-"end-of-line-hist",endoflinehist,ZLE_LINEMOVE,
-"exchange-point-and-mark",exchangepointandmark,ZLE_MOVE,
+"down-case-word",downcaseword,0,
+"down-history",downhistory,0,
+"down-line-or-history",downlineorhistory,ZLE_MOVEMENT|ZLE_LINEMOVE,
+"end-of-buffer-or-history",endofbufferorhistory,ZLE_MOVEMENT,
+"end-of-history",endofhistory,0,
+"end-of-line",endofline,ZLE_MOVEMENT,
+"end-of-line-hist",endoflinehist,ZLE_MOVEMENT,
+"exchange-point-and-mark",exchangepointandmark,ZLE_MOVEMENT,
 "execute-last-named-cmd",NULL,0,
 "execute-named-cmd",NULL,0,
-"expand-history",expandhistory,ZLE_MOD,
-"expand-or-complete",expandorcomplete,ZLE_MOD|ZLE_MENUCMP,
-"expand-word",expandword,ZLE_MOD,
-"forward-char",forwardchar,ZLE_MOVE,
-"forward-word",forwardword,ZLE_MOVE,
-"get-line",getline,ZLE_MOD,
-"gosmacs-transpose-chars",gosmacstransposechars,ZLE_MOD,
-"history-incremental-search-backward",historyincrementalsearchbackward,ZLE_MOD,
-"history-incremental-search-forward",historyincrementalsearchforward,ZLE_MOD,
-"history-search-backward",historysearchbackward,ZLE_MOD,
-"history-search-forward",historysearchforward,ZLE_MOD,
-"infer-next-history",infernexthistory,ZLE_MOD,
-"insert-last-word",insertlastword,ZLE_MOD,
-"kill-buffer",killbuffer,ZLE_MOD|ZLE_KILL,
-"kill-line",killline,ZLE_MOD|ZLE_KILL,
-"kill-region",killregion,ZLE_MOD|ZLE_KILL,
-"kill-whole-line",killwholeline,ZLE_MOD|ZLE_KILL,
+"expand-history",expandhistory,0,
+"expand-or-complete",expandorcomplete,ZLE_MENUCMP,
+"expand-word",expandword,0,
+"forward-char",forwardchar,ZLE_MOVEMENT,
+"forward-word",forwardword,ZLE_MOVEMENT,
+"get-line",getline,0,
+"gosmacs-transpose-chars",gosmacstransposechars,0,
+"history-incremental-search-backward",historyincrementalsearchbackward,0,
+"history-incremental-search-forward",historyincrementalsearchforward,0,
+"history-search-backward",historysearchbackward,ZLE_HISTSEARCH,
+"history-search-forward",historysearchforward,ZLE_HISTSEARCH,
+"infer-next-history",infernexthistory,0,
+"insert-last-word",insertlastword,0,
+"kill-buffer",killbuffer,ZLE_KILL,
+"kill-line",killline,ZLE_KILL,
+"kill-region",killregion,ZLE_KILL,
+"kill-whole-line",killwholeline,ZLE_KILL,
 "list-choices",listchoices,0,
 "list-expand",listexpand,ZLE_MENUCMP,
-"magic-space",magicspace,ZLE_MOD,
+"magic-space",magicspace,0,
 "menu-complete",menucompleteword,ZLE_MENUCMP,
 "menu-expand-or-complete",menuexpandorcomplete,ZLE_MENUCMP,
 "overwrite-mode",overwritemode,0,
-"push-line",pushline,ZLE_MOD,
-"quoted-insert",quotedinsert,ZLE_INSMOD,
-"quote-line",quoteline,ZLE_MOD,
-"quote-region",quoteregion,ZLE_MOD,
+"push-line",pushline,0,
+"quoted-insert",quotedinsert,0,
+"quote-line",quoteline,0,
+"quote-region",quoteregion,0,
 "redisplay",redisplay,0,
 "reverse-menu-complete",reversemenucomplete,ZLE_MENUCMP,
 "run-help",processcmd,0,
-"self-insert",selfinsert,ZLE_INSMOD,
-"self-insert-unmeta",selfinsertunmeta,ZLE_INSMOD,
+"self-insert",selfinsert,0,
+"self-insert-unmeta",selfinsertunmeta,0,
 "send-break",sendbreak,0,
 "send-string",sendstring,0,
 "",NULL,0,
 "set-mark-command",setmarkcommand,0,
-"spell-word",spellword,ZLE_MOD,
-"toggle-literal-history",toggleliteralhistory,ZLE_MOD,
-"transpose-chars",transposechars,ZLE_MOD,
-"transpose-words",transposewords,ZLE_MOD,
+"spell-word",spellword,0,
+"toggle-literal-history",toggleliteralhistory,0,
+"transpose-chars",transposechars,0,
+"transpose-words",transposewords,0,
 "undefined-key",undefinedkey,0,
 "undo",undo,ZLE_UNDO,
 "universal-argument",universalargument,ZLE_ARG,
-"up-case-word",upcaseword,ZLE_MOD,
-"up-history",uphistory,ZLE_MOD,
-"up-line-or-history",uplineorhistory,ZLE_LINEMOVE2,
+"up-case-word",upcaseword,0,
+"up-history",uphistory,0,
+"up-line-or-history",uplineorhistory,ZLE_LINEMOVE|ZLE_MOVEMENT,
 "vi-add-eol",viaddeol,0,
 "vi-add-next",viaddnext,0,
+"vi-backward-blank-word",vibackwardblankword,ZLE_MOVEMENT,
+"vi-backward-char",vibackwardchar,ZLE_MOVEMENT,
+"vi-backward-delete-char",vibackwarddeletechar,ZLE_KILL,
+"vi-beginning-of-line",vibeginningofline,ZLE_MOVEMENT,
 "vi-caps-lock-panic",vicapslockpanic,0,
-"vi-change",vichange,ZLE_MOD,
-"vi-change-eol",vichangeeol,ZLE_MOD,
-"vi-change-whole-line",vichangewholeline,ZLE_MOD,
+"vi-change",vichange,0,
+"vi-change-eol",vichangeeol,0,
+"vi-change-whole-line",vichangewholeline,0,
 "vi-cmd-mode",vicmdmode,0,
-"vi-delete",videlete,ZLE_MOD,
+"vi-delete",videlete,ZLE_KILL,
+"vi-delete-char",videletechar,ZLE_KILL,
 "vi-digit-or-beginning-of-line",NULL,0,
-"vi-fetch-history",vifetchhistory,ZLE_MOD,
-"vi-find-next-char",vifindnextchar,ZLE_MOVE,
-"vi-find-next-char-skip",vifindnextcharskip,ZLE_MOVE,
-"vi-find-prev-char",vifindprevchar,ZLE_MOVE,
-"vi-find-prev-char-skip",vifindprevcharskip,ZLE_MOVE,
-"vi-first-non-blank",vifirstnonblank,ZLE_MOVE,
-"vi-forward-word-end",viforwardwordend,ZLE_MOVE,
-"vi-goto-column",vigotocolumn,ZLE_MOVE,
-"vi-history-search-backward",vihistorysearchbackward,ZLE_MOD,
-"vi-history-search-forward",vihistorysearchforward,ZLE_MOD,
+"vi-end-of-line",viendofline,ZLE_MOVEMENT,
+"vi-fetch-history",vifetchhistory,0,
+"vi-find-next-char",vifindnextchar,ZLE_MOVEMENT,
+"vi-find-next-char-skip",vifindnextcharskip,ZLE_MOVEMENT,
+"vi-find-prev-char",vifindprevchar,ZLE_MOVEMENT,
+"vi-find-prev-char-skip",vifindprevcharskip,ZLE_MOVEMENT,
+"vi-first-non-blank",vifirstnonblank,ZLE_MOVEMENT,
+"vi-forward-blank-word",viforwardblankword,ZLE_MOVEMENT,
+"vi-forward-blank-word-end",viforwardblankwordend,ZLE_MOVEMENT,
+"vi-forward-char",viforwardchar,ZLE_MOVEMENT,
+"vi-forward-word-end",viforwardwordend,ZLE_MOVEMENT,
+"vi-goto-column",vigotocolumn,ZLE_MOVEMENT,
+"vi-goto-mark",vigotomark,ZLE_MOVEMENT,
+"vi-goto-mark-line",vigotomarkline,ZLE_MOVEMENT,
+"vi-history-search-backward",vihistorysearchbackward,0,
+"vi-history-search-forward",vihistorysearchforward,0,
+"vi-indent",viindent,0,
 "vi-insert",viinsert,0,
 "vi-insert-bol",viinsertbol,0,
-"vi-join",vijoin,ZLE_MOD,
-"vi-match-bracket",vimatchbracket,ZLE_MOVE,
-"vi-open-line-above",viopenlineabove,ZLE_MOD,
-"vi-open-line-below",viopenlinebelow,ZLE_MOD,
-"vi-oper-swap-case",vioperswapcase,ZLE_MOD,
-"vi-put-after",viputafter,ZLE_MOD,
-"vi-repeat-find",virepeatfind,ZLE_MOVE,
-"vi-repeat-search",virepeatsearch,ZLE_MOD,
-"vi-replace",vireplace,ZLE_MOD,
-"vi-replace-chars",vireplacechars,ZLE_MOD,
-"vi-rev-repeat-find",virevrepeatfind,ZLE_MOVE,
-"vi-rev-repeat-search",virevrepeatsearch,ZLE_MOD,
-"vi-swap-case",viswapcase,ZLE_MOD,
-"vi-undo",undo,ZLE_MOD,
+"vi-join",vijoin,0,
+"vi-match-bracket",vimatchbracket,ZLE_MOVEMENT,
+"vi-open-line-above",viopenlineabove,0,
+"vi-open-line-below",viopenlinebelow,0,
+"vi-oper-swap-case",vioperswapcase,0,
+"vi-put-after",viputafter,ZLE_YANK,
+"vi-repeat-change",virepeatchange,0,
+"vi-repeat-find",virepeatfind,ZLE_MOVEMENT,
+"vi-repeat-search",virepeatsearch,ZLE_MOVEMENT,
+"vi-replace",vireplace,0,
+"vi-replace-chars",vireplacechars,0,
+"vi-rev-repeat-find",virevrepeatfind,ZLE_MOVEMENT,
+"vi-rev-repeat-search",virevrepeatsearch,ZLE_MOVEMENT,
+"vi-set-buffer",visetbuffer,0,
+"vi-set-mark",visetmark,0,
+"vi-substitute",visubstitute,0,
+"vi-swap-case",viswapcase,0,
+"vi-undo-change",undo,0,
+"vi-unindent",viunindent,0,
 "vi-yank",viyank,0,
 "vi-yank-eol",viyankeol,0,
 "which-command",processcmd,0,
-"yank",yank,ZLE_MOD|ZLE_YANK,
-"yank-pop",yankpop,ZLE_MOD|ZLE_YANK,
+"yank",yank,ZLE_YANK|ZLE_NAMEDBUFFER,
+"yank-pop",yankpop,ZLE_YANK,
+"emacs-forward-word",emacsforwardword,ZLE_MOVEMENT,
+"emacs-backward-word",emacsbackwardword,ZLE_MOVEMENT,
+"kill-word",killword,ZLE_KILL,
+"vi-kill-line",vikillline,0,
+"vi-backward-delete-word",vibackwardkillword,0,
+"expand-cmd-path",expandcmdpath,0,
+"neg-argument",negargument,ZLE_NEGARG,
 "",NULL,0
 };
 
@@ -332,7 +356,7 @@ int emacsbind[256] = {
 /* M-* */ z_undefinedkey,
 /* M-+ */ z_undefinedkey,
 /* M-, */ z_undefinedkey,
-/* M-- */ z_undefinedkey,
+/* M-- */ z_negargument,
 /* M-. */ z_insertlastword,
 /* M-/ */ z_undefinedkey,
 /* M-0 */ z_digitargument,
@@ -426,7 +450,7 @@ int viinsbind[32] = {
 /* ^E */ z_selfinsert,
 /* ^F */ z_selfinsert,
 /* ^G */ z_selfinsert,
-/* ^H */ z_backwarddeletechar,
+/* ^H */ z_vibackwarddeletechar,
 /* ^I */ z_expandorcomplete,
 /* ^J */ z_acceptline,
 /* ^K */ z_killline,
@@ -439,9 +463,9 @@ int viinsbind[32] = {
 /* ^R */ z_redisplay,
 /* ^S */ z_selfinsert,
 /* ^T */ z_selfinsert,
-/* ^U */ z_killwholeline,
+/* ^U */ z_vikillline,
 /* ^V */ z_quotedinsert,
-/* ^W */ z_backwardkillword,
+/* ^W */ z_vibackwardkillword,
 /* ^X */ z_selfinsert,
 /* ^Y */ z_selfinsert,
 /* ^Z */ z_selfinsert,
@@ -485,21 +509,21 @@ int vicmdbind[128] = {
 /* ^] */ z_undefinedkey,
 /* ^^ */ z_undefinedkey,
 /* ^_ */ z_undefinedkey,
-/*   */ z_forwardchar,
+/*   */ z_viforwardchar,
 /* ! */ z_undefinedkey,
-/* " */ z_undefinedkey,
+/* " */ z_visetbuffer,
 /* # */ z_undefinedkey,
-/* $ */ z_endofline,
+/* $ */ z_viendofline,
 /* % */ z_vimatchbracket,
 /* & */ z_undefinedkey,
-/* ' */ z_undefinedkey,
+/* ' */ z_vigotomarkline,
 /* ( */ z_undefinedkey,
 /* ) */ z_undefinedkey,
 /* * */ z_undefinedkey,
 /* + */ z_downlineorhistory,
 /* , */ z_virevrepeatfind,
 /* - */ z_uplineorhistory,
-/* . */ z_undefinedkey,
+/* . */ z_virepeatchange,
 /* / */ z_vihistorysearchbackward,
 /* 0 */ z_vidigitorbeginningofline,
 /* 1 */ z_digitargument,
@@ -513,16 +537,16 @@ int vicmdbind[128] = {
 /* 9 */ z_digitargument,
 /* : */ z_undefinedkey,
 /* ; */ z_virepeatfind,
-/* < */ z_undefinedkey,
+/* < */ z_viunindent,
 /* = */ z_listchoices,
-/* > */ z_undefinedkey,
+/* > */ z_viindent,
 /* ? */ z_vihistorysearchforward,
 /* @ */ z_undefinedkey,
 /* A */ z_viaddeol,
-/* B */ z_undefinedkey,
+/* B */ z_vibackwardblankword,
 /* C */ z_vichangeeol,
 /* D */ z_killline,
-/* E */ z_undefinedkey,
+/* E */ z_viforwardblankwordend,
 /* F */ z_vifindprevchar,
 /* G */ z_vifetchhistory,
 /* H */ z_vicapslockpanic,
@@ -540,8 +564,8 @@ int vicmdbind[128] = {
 /* T */ z_vifindprevcharskip,
 /* U */ z_undefinedkey,
 /* V */ z_undefinedkey,
-/* W */ z_undefinedkey,
-/* X */ z_backwarddeletechar,
+/* W */ z_viforwardblankword,
+/* X */ z_vibackwarddeletechar,
 /* Y */ z_viyankeol,
 /* Z */ z_undefinedkey,
 /* [ */ z_undefinedkey,
@@ -549,7 +573,7 @@ int vicmdbind[128] = {
 /* ] */ z_undefinedkey,
 /* ^ */ z_vifirstnonblank,
 /* _ */ z_undefinedkey,
-/* ` */ z_undefinedkey,
+/* ` */ z_vigotomark,
 /* a */ z_viaddnext,
 /* b */ z_backwardword,
 /* c */ z_vichange,
@@ -557,12 +581,12 @@ int vicmdbind[128] = {
 /* e */ z_viforwardwordend,
 /* f */ z_vifindnextchar,
 /* g */ z_undefinedkey,
-/* h */ z_backwardchar,
+/* h */ z_vibackwardchar,
 /* i */ z_viinsert,
 /* j */ z_downlineorhistory,
 /* k */ z_uplineorhistory,
-/* l */ z_forwardchar,
-/* m */ z_undefinedkey,
+/* l */ z_viforwardchar,
+/* m */ z_visetmark,
 /* n */ z_virepeatsearch,
 /* o */ z_viopenlinebelow,
 /* p */ z_viputafter,
@@ -573,7 +597,7 @@ int vicmdbind[128] = {
 /* u */ z_undo,
 /* v */ z_undefinedkey,
 /* w */ z_forwardword,
-/* x */ z_deletechar,
+/* x */ z_videletechar,
 /* y */ z_viyank,
 /* z */ z_undefinedkey,
 /* { */ z_undefinedkey,
