@@ -32,8 +32,15 @@
 
 #define ZLE
 #include "zsh.h"
-#include "y.tab.h"
+#ifdef __hpux
+#include <ndir.h>
+#else
+#ifdef SYSV
+#define direct dirent
+#else
 #include <sys/dir.h>
+#endif
+#endif
 #include	<pwd.h>
 
 static int we,wb,usemenu,useglob;
@@ -42,13 +49,8 @@ static int menub,menue,menuw;
 static Lklist menulist;
 static Lknode menunode;
 
-#define INCMD (incmd||incond||inredir||incase)
-
 #define inststr(X) inststrlen((X),-1)
 
-/* strncpy() does not null always terminate the string, pem */
-#define STRNCPY(s1, s2, len)	(strncpy(s1, s2, len), s1[len] = '\0')
-
 int usetab() /**/
 {
 char *s = line+cs-1;
@@ -165,11 +167,10 @@ void acceptandmenucomplete() /**/
 {
 int t0,t1;
 
-	if (!menucmp)
-		{
+	if (!menucmp) {
 		feep();
 		return;
-		}
+	}
 	spaceinline(1);
 	line[cs++] = ' ';
 	spaceinline(menub-menuw);
@@ -181,138 +182,152 @@ int t0,t1;
 	menucompleteword();
 }
 
-#ifdef LASTMENU
 static char *lastmenu = NULL;
 static int lastmenupos = -1;
-#endif
+static int lincmd,lastambig;
+static char *cmdstr;
 
 void docomplete(lst) /**/
 int lst;
 {
-int t0,lincmd = INCMD;
 char *s;
 
-#ifdef LASTMENU
-	if (menucmp && lastmenu)
-		{
-		int     len = strlen(lastmenu);
-		if (lst >= COMP_EXPAND || cs != lastmenupos || strcmp(line, lastmenu) != 0)
-			{
-			free(lastmenu);
-			lastmenu = NULL;
-			lastmenupos = -1;
-			freemenu();
-			}
-		}
-#endif
-	if (menucmp)
-		{
-		if (lst == COMP_LIST_COMPLETE)
-			{
-			listmatches(menulist, NULL);
-			return;
-			}
-		cs = menub;
-		foredel(menue-menub);
-		incnode(menunode);
-		if (!menunode)
-			menunode = firstnode(menulist);
-		s = menunode->dat;
-		if (*s == '~' || *s == '=')
-			{
-			spaceinline(1);
-			line[cs++] = *s++;
-			}
-		inststr(s = menunode->dat);
-#ifdef LASTMENU
-		if (lastmenu)
-			free(lastmenu);
-		lastmenu = ztrdup(line);
-		lastmenupos = cs;
-#endif
-		menue = cs;
-		return;
-		}
-	if (doexpandhist())
-		return;
-	zleparse = 1;
-	eofseen = 0;
-	lexsave();
-	hungets(" "); /* KLUDGE! */
-	hungets(line);
-	strinbeg();
-	pushheap();
-	while (!eofseen && zleparse)
-		{
-		lincmd = INCMD;
-		if ((t0 = yylex()) == ENDINPUT)
-			break;
-		}
-	if (t0 == ENDINPUT)
-		{
-		s = ztrdup("");
-		we = wb = cs;
-		t0 = STRING;
-		}
-	else if (t0 == STRING)
-		s = ztrdup(yylval.str);
-	hflush();
-	strinend();
-	errflag = zleparse = 0;
-	if (we > ll)
-		we = ll;
-	if (lst == COMP_EXPAND_COMPLETE)
-		{
-		char *q = s;
+	if (isset(AUTOMENU) && !menucmp && c == '\t' &&
+		(lastcmd & ZLE_MENUCMP) && lastambig) usemenu = 1;
+	if (menucmp) { do_menucmp(lst); return; }
+	if (doexpandhist()) return;
+	s = get_comp_string();
+	if (s) {
+		if (lst == COMP_EXPAND_COMPLETE) {
+			char *q = s;
 
-		if (*q == Tilde)
-			lst = COMP_COMPLETE;
-		else if (*q == Equals)
-			{
-			for (q++; *q; q++)
-				if (*q == '/')
-					break;
-			if (!*q)
-				lst = COMP_EXPAND;
-			q = s+1;
+			if (*q == Tilde) q++;
+			else if (*q == Equals) {
+				for (q++; *q; q++)
+					if (*q == '/')
+						break;
+				if (!*q)
+					lst = COMP_EXPAND;
+				q = s+1;
+			} else {
+				for (; *q && *q != String; q++);
+				if (*q == String) {
+					if (getsparam(q+1)) lst = COMP_EXPAND;
+					else lst = COMP_COMPLETE;
+				}
+				q = s;
 			}
-		if (lst == COMP_EXPAND_COMPLETE)
-			{
-			for (; *q; q++)
-				if (itok(*q))
-					break;
-			if (!*q)
-				lst = COMP_COMPLETE;
+			if (lst == COMP_EXPAND_COMPLETE) {
+				for (; *q; q++)
+					if (itok(*q))
+						break;
+				if (!*q)
+					lst = COMP_COMPLETE;
 			}
 		}
-	if (t0 != STRING)
-		feep();
-	else
-		{
-		if (lst == COMP_SPELL)
-			{
+		if (lst == COMP_SPELL) {
 			char	**x = &s;
 			untokenize(s);
 			cs = wb;
 			foredel(we-wb);
 			/* call the real spell checker, ash@aaii.oz.zu */
-			spckword(x, NULL, lincmd, 0);
+			spckword(x, NULL, NULL, lincmd, 0);
 			inststr(*x);
-			}
-		else if (COMP_ISEXPAND(lst))
+		} else if (COMP_ISEXPAND(lst))
 			doexpansion(s,lst,lincmd);
-		else
-			{
-			if (!useglob)
-				untokenize(s);
+		else {
 			docompletion(s,lst,lincmd);
-			}
-		free(s);
 		}
+		free(s);
+	}
 	popheap();
 	lexrestore();
 }
 
+void do_menucmp(lst) /**/
+int lst;
+{
+char *s;
+
+	if (isset(LASTMENU) && lastmenu) {
+		if (COMP_ISEXPAND(lst) || cs != lastmenupos ||
+				strcmp(line, lastmenu) != 0) {
+			free(lastmenu);
+			lastmenu = NULL;
+			lastmenupos = -1;
+			freemenu();
+		}
+	}
+	if (lst == COMP_LIST_COMPLETE) {
+		listmatches(menulist, NULL);
+		return;
+	}
+	cs = menub;
+	foredel(menue-menub);
+	incnode(menunode);
+	if (!menunode)
+		menunode = firstnode(menulist);
+	s = menunode->dat;
+	if (*s == '~' || *s == '=' || *s == '$') {
+		spaceinline(1);
+		line[cs++] = *s++;
+	}
+	inststr(s = menunode->dat);
+	if (isset(LASTMENU)) {
+		if (lastmenu) free(lastmenu);
+		lastmenu = ztrdup(line);
+		lastmenupos = cs;
+	}
+	menue = cs;
+}
+
+char *get_comp_string() /**/
+{
+int t0;
+char *s,*linptr;
+
+	linptr = line;
+start:
+	lincmd = incmdpos;
+	cmdstr = NULL;
+	zleparse = 1;
+	lexsave();
+	hungets(" "); /* KLUDGE! */
+	hungets(linptr);
+	strinbeg();
+	pushheap();
+	do {
+		lincmd = incmdpos;
+		ctxtlex();
+		if (tok == ENDINPUT) break;
+		if (lincmd) cmdstr = strdup(tokstr);
+	} while (tok != ENDINPUT && zleparse);
+	t0 = tok;
+	if (t0 == ENDINPUT) {
+		s = ztrdup("");
+		we = wb = cs;
+		t0 = STRING;
+	} else if (t0 == STRING) {
+		s = ztrdup(tokstr);
+	} else if (t0 == ENVSTRING) {
+		for (s = tokstr; *s && *s != '='; s++, wb++);
+		if (*s) { s++; wb++; t0 = STRING; s = ztrdup(s); }
+		lincmd = 1;
+	}
+	hflush();
+	strinend();
+	errflag = zleparse = 0;
+	if (we > ll) we = ll;
+	if (t0 == LEXERR && parbegin != -1) {
+		linptr += ll+1-parbegin;
+		popheap();
+		lexrestore();
+		goto start;
+	}
+	if (t0 != STRING) { feep(); return NULL; }
+	return s;
+}
+
 void doexpansion(s,lst,lincmd) /**/
 char *s;int lst;int lincmd;
 {
@@ -327,39 +342,30 @@ char *ss;
 	postfork(vl,1);
 	if (errflag)
 		goto end;
-	if (lst == COMP_LIST_EXPAND)
-		{
-		listmatches(vl,NULL);
-		goto end;
-		}
-	else if (peekfirst(vl) == s) 
-		{
-		if (lst == COMP_EXPAND_COMPLETE)
-			{
-			untokenize(s);
-			docompletion(s,COMP_COMPLETE,lincmd);
-			}
-		else
-			feep();
-		goto end;
-		}
-	if (!full(vl) || !*(char *) peekfirst(vl))
-		{
+	if (!full(vl) || !*(char *) peekfirst(vl)) {
 		feep();
 		goto end;
-		}
+	}
+	if (lst == COMP_LIST_EXPAND) {
+		listmatches(vl,NULL);
+		goto end;
+	} else if (peekfirst(vl) == s) {
+		if (lst == COMP_EXPAND_COMPLETE) {
+			docompletion(s,COMP_COMPLETE,lincmd);
+		} else
+			feep();
+		goto end;
+	}
 	cs = wb;
 	foredel(we-wb);
-	while (ss = ugetnode(vl))
-		{
+	while (ss = ugetnode(vl)) {
 		untokenize(ss);
 		inststr(ss);
-		if (full(vl))
-			{
+		if (full(vl)) {
 			spaceinline(1);
 			line[cs++] = ' ';
-			}
 		}
+	}
 end:
 	popheap();
 	setterm();
@@ -412,6 +418,7 @@ char *t,*u,*v;
 static int ambig,haspath,exact;
 static Lklist matches;
 static char *pat;
+static int typechar;
 
 void addmatch(s) /**/
 char *s;
@@ -466,19 +473,29 @@ char	*s;
 }
 
 /*
- * opendir that handles '~' and '='.
+ * opendir that handles '~' and '=' and '$'.
  * orig. by ash@aaii.oz.au, mod. by pf
  */
 DIR *OPENDIR(s)
 char	*s;
 {
-	if (*s != '~' && *s != '=')
+	if (*s != '~' && *s != '=' && *s != '$')
 		return(opendir(s));
 	s = strdup(s);
-	*s = (*s == '=') ? Equals : Tilde;
-	filesub(&s);
+	*s = (*s == '=') ? Equals : (*s == '~') ? Tilde : String;
+	singsub(&s);
 	return(opendir(s));
 }
+char *dirname(s)
+char	*s;
+{
+	if (*s == '~' || *s == '=' || *s == '$') {
+	  s = strdup(s);
+	  *s = (*s == '=') ? Equals : (*s == '~') ? Tilde : String;
+	  singsub(&s);
+	}
+	return(s);
+}
 
 int Isdir(s) /**/
 char *s;
@@ -496,95 +513,153 @@ char *t;char *s;
 {
 char buf[MAXPATHLEN];
 
-	sprintf(buf,"%s/%s",(s) ? s : ".",t);
+	if (typechar != '$')
+		sprintf(buf,"%s/%s",(s) ? s : ".",t);
+	else
+		sprintf(buf,"$%s",t);
 	s = buf;
-	if (*s != '~' && *s != '=' && *s != Tilde && *s != Equals)
+	if (*s != '~' && *s != '=' && *s != Tilde && *s != Equals &&
+		 *s != '$' && *s != String)
 		return(Isdir(s));
 	s = strdup(s);
-	if (*s == '~' || *s == '=')
-		*s = (*s == '=') ? Equals : Tilde;
-	filesub(&s);
+	if (*s == '~' || *s == '=' || *s == '$')
+		*s = (*s == '=') ? Equals : (*s == '~') ? Tilde : String;
+	singsub(&s);
 	return(Isdir(s));
 }
 
+#define SLASH_YES   0
+#define SLASH_NO    1
+#define SLASH_MAYBE 2
+
+int slashflag;
+int addedstar;
+char *pathprefix;
+
 void docompletion(s,lst,incmd) /**/
 char *s;int lst;int incmd;
 {
-DIR *d;
-struct direct *de;
-char *u,*ppfx = NULL;
-int commcomp = 0;
-int tildeexpand = 0;
-char	*ss;
-int addedstar = 0;
+char *tokorigs;
+char *origs;
+
+	slashflag = SLASH_MAYBE;
+	addedstar = 0;
+	lastambig = 0;
 
 	heapalloc();
 	pushheap();
-	ss = strdup(s);
-	matches = newlist();
 	if (useglob)
-		{
-		char	*pt;
-		char	ch;
-		int	hasp = 0;
+	    tokorigs = strdup(s);
+	untokenize(s);
+	origs = strdup(s);
+	matches = newlist();
+	if (cmdstr && inarray(cmdstr,hostcmds)) {
+		char **x;
+		haspath = exact = 0;
+		slashflag = SLASH_NO;
+		pat = s;
+		for (x = hosts; *x; x++) addcmdmatch(*x,NULL);
+	} else if (cmdstr && inarray(cmdstr,optcmds)) {
+		struct option *o;
+		haspath = exact = 0;
+		slashflag = SLASH_NO;
+		pat = s;
+		for (o = optns; o->name; o++) addcmdmatch(o->name,NULL);
+	} else if (cmdstr && inarray(cmdstr,varcmds)) {
+		haspath = exact = 0;
+		slashflag = SLASH_NO;
+		pat = s;
+		listhtable(paramtab,addcmdmatch);
+	} else if (cmdstr && inarray(cmdstr,bindcmds)) {
+		int t0;
+		haspath = exact = 0;
+		slashflag = SLASH_NO;
+		pat = s;
+		for (t0 = 0; t0 != ZLECMDCOUNT; t0++)
+			if (*zlecmds[t0].name) addcmdmatch(zlecmds[t0].name,NULL);
+	} else {
+		gen_matches_reg(s,incmd);
+		/* only do "globbed" completion if regular completion fails.
+		   pem, 7Oct91 */
+		if ((!full(matches) || errflag) && useglob) {
+		    gen_matches_glob(tokorigs,incmd);
+		    /*
+		     * gen_matches_glob changes the insert line to be correct up
+		     * to the match, so the prefix string must be "". ash, 7Oct91
+		     */
+		    *s = 0;
+		}
+	}
+	if (lst != COMP_LIST_COMPLETE) do_fignore(origs);
+	if (!full(matches) || errflag) {
+		feep();
+	} else if (lst == COMP_LIST_COMPLETE) {
+		listmatches(matches,
+			unset(LISTTYPES) ? NULL :
+				(haspath) ? pathprefix : "./");
+	} else if (nextnode(firstnode(matches))) {
+		do_ambiguous(s);
+	} else {
+		do_single(s);
+	}
+	ll = strlen(line);
+	setterm();
+	popheap();
+	permalloc();
+}
 
-		/*
-		 * Find the longest prefix string without any
-		 * chars special to glob - ash.
-		 */
-		for (pt = s; *pt; pt++)
-		    {
-		    if (pt == s && (*pt == Tilde || *pt == Equals))
-			continue;
-		    if (ispecial(*pt) || itok(*pt))
-			break;
-		    }
-		for (; pt > s && *pt != '/'; pt--)
-			;
-		if (*pt == '/')
-		    {
-		    *pt = 0;
-		    u = pt + 1;
-		    wb += strlen(s) + 1;
-		    hasp = 1;
-		    }
-		else
-		    u = s;
-		if (!hasp && (*s == Tilde || *s == Equals)) 
-			{
-			/* string contains only ~xx, so do tilde expansion */
-			maketildelist(s);
-			wb++;
-			ppfx = s;
-			tildeexpand = 1;
-			}
-		else if (commcomp = !incmd && !hasp)
-			{
-			pat = s;
-			listhtable(aliastab ,addcmdmatch);
-			listhtable(cmdnamtab,addcmdnodis);
-			if (d = opendir("."))
-				{
-				char *q;
+void gen_matches_glob(s,incmd) /**/
+char *s;int incmd;
+{
+char *pt,*u;
+int hasp = 0;
+DIR *d;
+struct direct *de;
 
-				readdir(d); readdir(d);
-				while ((de = readdir(d)) && !errflag)
-					if (strpfx(pat,q = de->d_name) &&
-								(*q != '.' || *u == '.' || isset(GLOBDOTS)))
-							addmatch(q);
-				closedir(d);
-				}
-			}
-		else
-		    {
-		    int 	commonprefix = 0;
-		    char	*prefix;
-		    Lknode	n;
-		    int		nonomatch = isset(NONOMATCH);
+	/*
+	 * Find the longest prefix string without any
+	 * chars special to glob - ash.
+	 */
+	for (pt = s; *pt; pt++) {
+		if (pt == s && (*pt == Tilde || *pt == Equals)) continue;
+		if (ispecial(*pt) || itok(*pt)) break;
+	}
+	for (; pt > s && *pt != '/'; pt--) ;
+	if (*pt == '/') {
+		*pt = 0;
+		u = pt + 1;
+		wb += strlen(s);
+		hasp = 1;
+	} else u = s;
+	if (!hasp && (*s == Tilde || *s == Equals)) {
+		/* string contains only ~xx, so do tilde expansion */
+		maketildelist(s);
+		wb++;
+		pathprefix = s;
+		slashflag = SLASH_YES;
+	} else if (incmd && !hasp) {
+		slashflag = SLASH_NO;
+		pat = s;
+		listhtable(aliastab ,addcmdmatch);
+		listhtable(cmdnamtab,addcmdnodis);
+		if (d = opendir(".")) {
+			char *q;
 
-		    opts[NONOMATCH] = 1;
-		    if (hasp)
-			{
+			readdir(d); readdir(d);
+			while ((de = readdir(d)) && !errflag)
+				if (strpfx(pat,q = de->d_name) &&
+							(*q != '.' || *u == '.' || isset(GLOBDOTS)))
+					addmatch(q);
+			closedir(d);
+		}
+	} else {
+		 int 	commonprefix = 0;
+		 char	*prefix;
+		 Lknode	n;
+		 int		nonomatch = isset(NONOMATCH);
+
+		 opts[NONOMATCH] = 1;
+		 if (hasp) {
 			/* Find the longest common prefix string
 			 * after globbing the input. All expansions
 			 * ~foo/bar/* will turn into something like
@@ -592,259 +667,263 @@ int addedstar = 0;
 			 * We will remove this common prefix from the matches.
 			 * ash, 7 May '91
 			 */
-			ppfx = s;
+			pathprefix = s;
 			addnode(matches,s);
 			prefork(matches);
-			if (!errflag)
-				postfork(matches,1);
-			if (!errflag)
-				{
+			if (!errflag) postfork(matches,1);
+			if (!errflag) {
 				prefix = peekfirst(matches);
-				if (prefix)
-				    commonprefix = strlen(prefix) + 1;
+				if (prefix) commonprefix = strlen(prefix) + 1;
 				*pt = '/';
-				}
 			}
-		    if (s[strlen(s) - 1] == '/') {
+		}
+		if (s[strlen(s) - 1] == '/') {
 			/* if strings ends in a '/' always add a '*' */
 			s = dyncat(s,"x");
 			s[strlen(s)-1] = Star;
 			addedstar = 1;
-		    };
-		    matches = newlist();
-		    addnode(matches,s);
-		    prefork(matches);
-		    if (!errflag)
-			    postfork(matches,1);
-		    opts[NONOMATCH] = nonomatch;
-		    if (errflag || !full(matches) || !nextnode(firstnode(matches)))
-			    {
-			    /* if there were no matches (or only one)
-			       add a trailing * and try again */
-			    s = dyncat(s,"x");
-			    s[strlen(s)-1] = Star;
-			    addedstar = 1;
-			    matches = newlist();
-			    addnode(matches,s);
-			    prefork(matches);
-			    if (errflag)
-				    goto end;
-			    postfork(matches,1);
-			    if (errflag)
-				    goto end;
-			    }
-		    if (commonprefix)
-			/* remove the common prefix from all the matches */
+		}
+		matches = newlist();
+		addnode(matches,s);
+		prefork(matches);
+		if (!errflag) postfork(matches,1);
+		opts[NONOMATCH] = nonomatch;
+		if (errflag || !full(matches) || !nextnode(firstnode(matches))) {
+			/* if there were no matches (or only one)
+				add a trailing * and try again */
+			s = dyncat(s,"x");
+			s[strlen(s)-1] = Star;
+			addedstar = 1;
+			matches = newlist();
+			addnode(matches,s);
+			prefork(matches);
+			if (errflag) return;
+			postfork(matches,1);
+			if (errflag) return;
+		}
+		/* remove the common prefix from all the matches */
+		if (commonprefix)
 			for (n = firstnode(matches); n; incnode(n))
 				n->dat = (char *) n->dat+commonprefix;
-		    s = pt;
-		    *s = 0;
-		    }
-  		}
-	else
-		{
-		haspath = exact = 0;
-		for (u = s+strlen(s); u >= s && *u != '/'; u--);
-		if (u >= s)
-			{
-			*u++ = '\0';
-			haspath = 1;
-			}
-		else
-			u = s;
-		pat = u;
-		if (commcomp = !incmd && !haspath)
-			{
-			listhtable(aliastab ,addcmdmatch);
-			listhtable(cmdnamtab,addcmdnodis);
-			if (d = opendir("."))
-				{
-				char *q;
+		s = pt;
+		*s = 0;
+	}
+}
 
-				readdir(d); readdir(d);
-				while ((de = readdir(d)) && !errflag)
-					if (strpfx(pat,q = de->d_name) &&
-								(*q != '.' || *u == '.' || isset(GLOBDOTS)))
-							addmatch(q);
-				closedir(d);
-				}
-			}
-		else if ((*s == '~' || *s == '=') && !haspath)
-			{
-			maketildelist(s);
-			ppfx = s;
-			tildeexpand = 1;
-			}
-		else if (d = OPENDIR(ppfx =
-				((haspath || *s == '~') ? ((*s) ? s : "/") : ".")))
-			{
+void gen_matches_reg(s,incmd) /**/
+char *s;int incmd;
+{
+char *u;
+DIR *d;
+struct direct *de;
+
+	haspath = exact = 0;
+	for (u = s+strlen(s); u >= s; u--)
+		if (*u == '/' || *u == '@' || *u == '$') break;
+	typechar = *u;
+	if (u >= s) {
+		*u++ = '\0';
+		haspath = 1;
+	} else u = s;
+	pat = u;
+	if (typechar == '$' && haspath) {
+		/* slashflag = SLASH_NO; */
+		listhtable(paramtab,addcmdmatch);
+	} else if (typechar == '@' && haspath) {
+		char **x;
+		slashflag = SLASH_NO;
+		for (x = hosts; *x; x++) addcmdmatch(*x,NULL);
+	} else if (*s == '~' && !haspath) {
+		maketildelist(s);
+		pathprefix = s;
+		slashflag = SLASH_YES;
+	} else if (incmd && !haspath) {
+		slashflag = SLASH_NO;
+		listhtable(aliastab ,addcmdmatch);
+		listhtable(cmdnamtab,addcmdnodis);
+		if (d = opendir(".")) {
 			char *q;
+			struct stat buf;
 
 			readdir(d); readdir(d);
 			while ((de = readdir(d)) && !errflag)
 				if (strpfx(pat,q = de->d_name) &&
-							(*q != '.' || *u == '.' || isset(GLOBDOTS)))
-						addmatch(q);
+						(*q != '.' || *u == '.' || isset(GLOBDOTS)) &&
+						stat(q,&buf) >= 0 &&
+						(buf.st_mode & (S_IFMT|S_IEXEC)) == (S_IFREG|S_IEXEC))
+					addmatch(q);
 			closedir(d);
-			}
 		}
-	if (full(matches) && nextnode(firstnode(matches)))
-		{
+	} else if (d = OPENDIR(pathprefix =
+			((haspath || *s == '~') ? ((*s) ? s : "/") : "."))) {
+		char *q,buf2[MAXPATHLEN];
+		struct stat buf;
+		char dn[MAXPATHLEN];
+		
+		strcpy(dn,dirname(pathprefix));
+		readdir(d); readdir(d);
+		while ((de = readdir(d)) && !errflag)
+		  if (strpfx(pat,q = de->d_name) &&
+				(*q != '.' || *u == '.' || isset(GLOBDOTS))) {
+			 if (incmd) {
+				sprintf(buf2,"%s/%s",dn,q);
+				if (stat(buf2,&buf) < 0 ||
+					 (buf.st_mode & S_IEXEC) == S_IEXEC) {
+				  addmatch(q);
+				}
+			 } else {
+				addmatch(q);
+			 }
+		  }
+		closedir(d);
+	}
+}
+
+void do_fignore(origstr) /**/
+char *origstr;
+{
+	if (full(matches) && nextnode(firstnode(matches))) {
 		Lknode z,zn;
 
-		for (z = firstnode(matches); z; z = zn)
-			{
+		for (z = firstnode(matches); z; z = zn) {
 			char *q = getdata(z);
 			int namlen = strlen(q);
-			int	slen = strlen(ss);
+			int	slen = strlen(origstr);
 			char **pt = fignore;
 	
 			zn = nextnode(z);
-			for (; *pt; pt++)
-				{
+			for (; *pt; pt++) {
 				/* We try to be smart here and override the
 				   fignore variable if the user has explicity
 				   used the ignored prefix, pem, 7 May 1991 */
-				if (!addedstar && strcmp(ss+slen-strlen(*pt), *pt) == 0)
+				if (!addedstar && strcmp(origstr+slen-strlen(*pt), *pt) == 0)
 					continue;
-				if (strlen(*pt) < namlen && !strcmp(q+namlen-strlen(*pt),*pt))
-					{
+				if (strlen(*pt) < namlen && !strcmp(q+namlen-strlen(*pt),*pt)) {
 					uremnode(matches,z);
 					break;
-					}
 				}
 			}
 		}
-	if (!full(matches) || errflag)
+	}
+}
+
+void do_ambiguous(s) /**/
+char *s;
+{
+	lastambig = 1;
+	if (usemenu) { do_ambig_menu(s); return; }
+	if (useglob) {
 		feep();
-	else if (lst == COMP_LIST_COMPLETE)
-		listmatches(matches,
-					unset(LISTTYPES) ? NULL : (haspath) ? ppfx : "./");
-	else if (nextnode(firstnode(matches)))
-		{
-		if (usemenu)
-			{
-			menucmp = 1;
-			if (isset(MENUCOMPLETEBEEP))
-				feep();
-			cs = wb;
-			menuw = cs;
-			foredel(we-wb);
-			if (*s == '~' || *s == '=')
-				{
-				spaceinline(1);
-				line[cs++] = *s++;
-				}
-			if (haspath)
-				{
-				inststr(s);
-				spaceinline(1);
-				line[cs++] = '/';
-				}
-			menub = cs;
-			s = peekfirst(matches);
-			if ((*s == '~' || *s == '=') && !haspath)
-				{
-				spaceinline(1);
-				line[cs++] = *s++;
-				}
-			inststr(s);
-			menue = cs;
-			permalloc();
-			menulist = duplist(matches,ztrdup);
-			heapalloc();
-			menunode = firstnode(menulist);
-			popheap();
-			permalloc();
-#ifdef LASTMENU
-			if (lastmenu)
-				free(lastmenu);
-			lastmenu = ztrdup(line);
-			lastmenupos = cs;
-#endif
-			return;
-			}
-		else if (useglob)
-			{
-			feep();
-			if (isset(AUTOLIST))
-				listmatches(matches,
-					unset(LISTTYPES) ? NULL : (haspath) ? ppfx : "./");
-			goto end;
-			}
-		cs = wb;
-		foredel(we-wb);
-		if (*s == '~' || *s == '=')
-			{
-			spaceinline(1);
-			line[cs++] = *s++;
-			}
-		if (haspath)
-			{
-			inststr(s);
-			spaceinline(1);
-			line[cs++] = '/';
-			}
-		if (isset(RECEXACT) && exact)
-			{
-			if ((*u == '~' || *u == '=') && !haspath)
-				{
-				spaceinline(1);
-				line[cs++] = *s++;
-				}
-			inststr(u);
-			spaceinline(1);
-			line[cs++] = (!commcomp && (tildeexpand ||
-				isdir(u,ppfx)) ? '/' : ' ');
-			}
-		else
-			{
-			s = peekfirst(matches);
-			if ((*s == '~' || *s == '=') && !haspath)
-				{
-				spaceinline(1);
-				line[cs++] = *s++;
-				ambig--;
-				}
-			inststrlen(s,ambig);
-			refresh();
-			feep();
-			if (isset(AUTOLIST))
-				listmatches(matches,
-					unset(LISTTYPES) ? NULL : (haspath) ? ppfx : "./");
-			}
-		}
-	else
-		{
-		cs = wb;
-		foredel(we-wb);
-		if (*s == '~' || *s == '=')
-			{
-			spaceinline(1);
-			line[cs++] = *s++;
-			}
-		if (haspath)
-			{
-			inststr(s);
-			spaceinline(1);
-			line[cs++] = '/';
-			}
-		s = peekfirst(matches);
-		if ((*s == '~' || *s == '=') && !haspath)
-			{
-			spaceinline(1);
-			line[cs++] = *s++;
-			}
+		if (isset(AUTOLIST))
+			listmatches(matches,
+				unset(LISTTYPES) ? NULL : (haspath) ? pathprefix : "./");
+		return;
+	}
+	cs = wb;
+	foredel(we-wb);
+	if (*s == '~' || *s == '=' || *s == '$') {
+		spaceinline(1);
+		line[cs++] = *s++;
+	}
+	if (haspath) {
 		inststr(s);
 		spaceinline(1);
-		line[cs++] = (!commcomp && (tildeexpand ||
-			isdir(peekfirst(matches),ppfx)) ? '/' : ' '); 
+		line[cs++] = typechar;
+	}
+	if (isset(RECEXACT) && exact) {
+		lastambig = 0;
+		if ((*pat == '~' || *pat == '=' || *pat == '$') && !haspath) {
+			spaceinline(1);
+			line[cs++] = *s++;
 		}
-	ll = strlen(line);
-end:
-	setterm();
-	popheap();
+		inststr(pat);
+		spaceinline(1);
+		switch (slashflag) {
+			case SLASH_YES: line[cs++] = '/'; break;
+			case SLASH_NO : line[cs++] = ' '; break;
+			case SLASH_MAYBE: line[cs++] = isdir(pat,pathprefix)?'/':' '; break;
+		}
+		return;
+	}
+	s = peekfirst(matches);
+	if ((*s == '~' || *s == '=' || *s == '$') && !haspath) {
+		spaceinline(1);
+		line[cs++] = *s++;
+		ambig--;
+	}
+	inststrlen(s,ambig);
+	refresh();
+	if (isset(AUTOLIST)) {
+		if (unset(NOLISTBEEP)) feep();
+		listmatches(matches,
+			unset(LISTTYPES) ? NULL : (haspath) ? pathprefix : "./");
+	} else feep();
+}
+
+void do_single(s) /**/
+char *s;
+{
+	cs = wb;
+	foredel(we-wb);
+	if (*s == '~' || *s == '=' || *s == '$') {
+		spaceinline(1);
+		line[cs++] = *s++;
+	}
+	if (haspath) {
+		inststr(s);
+		spaceinline(1);
+		line[cs++] = typechar;
+	}
+	s = peekfirst(matches);
+	if ((*s == '~' || *s == '=' || *s == '$') && !haspath) {
+		spaceinline(1);
+		line[cs++] = *s++;
+	}
+	inststr(s);
+	spaceinline(1);
+	switch (slashflag) {
+		case SLASH_YES: line[cs++] = '/'; break;
+		case SLASH_NO : line[cs++] = ' '; break;
+		case SLASH_MAYBE: line[cs++] = isdir(s,pathprefix) ? '/' : ' '; break;
+	}
+}
+
+void do_ambig_menu(s) /**/
+char *s;
+{
+	menucmp = 1;
+	if (isset(MENUCOMPLETEBEEP)) feep();
+	cs = wb;
+	menuw = cs;
+	foredel(we-wb);
+	if (*s == '~' || *s == '=' || *s == '$') {
+		spaceinline(1);
+		line[cs++] = *s++;
+	}
+	if (haspath) {
+		inststr(s);
+		spaceinline(1);
+		line[cs++] = typechar;
+	}
+	menub = cs;
+	s = peekfirst(matches);
+	if ((*s == '~' || *s == '=' || *s == '$') && !haspath) {
+		spaceinline(1);
+		line[cs++] = *s++;
+	}
+	inststr(s);
+	menue = cs;
 	permalloc();
+	menulist = duplist(matches,(VFunc)ztrdup);
+	heapalloc();
+	menunode = firstnode(menulist);
+	permalloc();
+	if (isset(LASTMENU)) {
+		if (lastmenu)
+			free(lastmenu);
+		lastmenu = ztrdup(line);
+		lastmenupos = cs;
+	}
 }
 
 int strpfx(s,t) /**/
@@ -893,7 +972,9 @@ char **arr,**ap;
 			*apps = Tilde;
 		else if (*apps == '=')
 			*apps = Equals;
-		filesub(&apps);
+		else if (*apps == '$')
+			*apps = String;
+		singsub(&apps);
 		longest++;
 		}
 	qsort(arr,ct,sizeof(char *),forstrcmp);
@@ -907,6 +988,7 @@ char **arr,**ap;
 		{
 		ap = arr+t1;
 		if (apps)
+			{
 			do
 				{
 				int t2 = strlen(*ap)+1;
@@ -925,8 +1007,13 @@ char **arr,**ap;
 #endif
 					case S_IFCHR: putchar('%'); break;
 					case S_IFBLK: putchar('#'); break;
-					case S_IFLNK: putchar('@'); break;
+#ifdef S_IFLNK
+					case S_IFLNK: putchar(
+						(access(pbuf,F_OK) == -1) ? '&' : '@'); break;
+#endif
+#ifdef S_IFSOCK
 					case S_IFSOCK: putchar('='); break;
+#endif
 					default:
 						if (buf.st_mode & 0111)
 							putchar('*');
@@ -938,6 +1025,7 @@ char **arr,**ap;
 				for (t0 = colsz; t0 && *ap; t0--,ap++);
 				}
 			while (*ap);
+			}
 		else
 			do
 				{
@@ -1003,15 +1091,19 @@ char **arr,**ap;
 int doexpandhist() /**/
 {
 char *cc,*ce;
-int t0;
+int t0,oldcs,oldll;
 
 	for (cc = line, ce = line+ll; cc < ce; cc++)
-		if (*cc == bangchar || *cc == hatchar)
+		if (*cc == '\\' && cc[1])
+			cc++;
+		else if (*cc == bangchar)
 			break;
-	if (cc == ce)
+	if (*cc == bangchar && cc[1] == '"') return 0;
+	if (cc == ce && *line != hatchar)
 		return 0;
+	oldcs = cs;
+	oldll = ll;
 	zleparse = 1;
-	eofseen = 0;
 	lexsave();
 	hungets(line);
 	strinbeg();
@@ -1032,6 +1124,7 @@ int t0;
 	t0 = histdone;
 	lexrestore();
 	line[ll = cs] = '\0';
+	if (ll == oldll) cs = oldcs;
 	return t0;
 }
 
@@ -1048,30 +1141,30 @@ void expandhistory() /**/
 		feep();
 }
 
+static int cmdwb,cmdwe;
+
 char *getcurcmd() /**/
 {
-int t0,lincmd = INCMD;
+int lincmd = incmdpos;
 char *s = NULL;
 
 	zleparse = 1;
-	eofseen = 0;
 	lexsave();
 	hungets(" "); /* KLUDGE! */
 	hungets(line);
 	strinbeg();
 	pushheap();
-	while (!eofseen && zleparse)
-		{
-		if ((t0 = yylex()) == ENDINPUT)
-			break;
-		else if (t0 == STRING && !lincmd)
-			{
-			if (s)
-				free(s);
-			s = ztrdup(yylval.str);
-			}
-		lincmd = INCMD;
+	do {
+		lincmd = incmdpos;
+		ctxtlex();
+		if (tok == ENDINPUT) break;
+		if (tok == STRING && lincmd) {
+			if (s) free(s);
+			s = ztrdup(tokstr);
+			cmdwb = ll-wordbeg; cmdwe = ll+1-inbufct;
 		}
+		lincmd = incmdpos;
+	} while (tok != ENDINPUT && zleparse);
 	hflush();
 	popheap();
 	strinend();
@@ -1082,37 +1175,54 @@ char *s = NULL;
 
 void processcmd() /**/
 {
-char *s;
+char *s,*t;
 
 	s = getcurcmd();
-	if (s)
-		{
-		char *t;
+	if (!s) { feep(); return; }
+	t = zlecmds[bindk].name;
+	mult = 1;
+	pushline();
+	sizeline(strlen(s)+strlen(t)+1);
+	strcpy(line,t);
+	strcat(line," ");
+	cs = ll = strlen(line);
+	inststr(s);
+	free(s);
+	done = 1;
+}
 
-		t = zlecmds[bindk].name;
-		mult = 1;
-		pushline();
-		sizeline(strlen(s)+strlen(t)+1);
-		strcpy(line,t);
-		strcat(line," ");
-		cs = ll = strlen(line);
-		inststr(s);
-		free(s);
-		done = 1;
-		}
-	else
-		feep();
+void expandcmdpath() /**/
+{
+int oldcs = cs;
+char *s,*str;
+
+	s = getcurcmd();
+	if (!s) { feep(); return; }
+	str = findcmd(s);
+	free(s);
+	if (!str) { feep(); return; }
+	cs = cmdwb;
+	foredel(cmdwe-cmdwb);
+	spaceinline(strlen(str));
+	strncpy(line+cs,str,strlen(str));
+	cs = oldcs;
+	if (cs >= cmdwe) cs += cmdwe-cmdwb+strlen(str);
+	if (cs > ll) cs = ll;
+	free(str);
 }
 
 void freemenu() /**/
 {
-	if (menucmp)
-#ifdef LASTMENU
-	    if (lastmenu == NULL)
-#endif
-		{
+	if (menucmp && (unset(LASTMENU) || lastmenu == NULL)) {
 		menucmp = 0;
 		freetable(menulist,freestr);
-		}
+	}
+}
+
+int inarray(s,a) /**/
+char *s; char **a;
+{
+	for (; *a; a++) if (!strcmp(*a,s)) return 1;
+	return 0;
 }
 
