@@ -31,13 +31,31 @@
 */
 
 #include "config.h"
+
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
+
+#ifdef SYSV
+#include <sys/bsdtypes.h>
+#define _POSIX_SOURCE
+#include <sys/limits.h>
+#include <sys/sioctl.h>
+#define MAXPATHLEN PATH_MAX
+#define lstat stat
+extern int gethostname();
+#define sigmask(m) m
+#include <sys/dirent.h>
+struct rusage { int foo; }
+#else
 #include <sys/types.h>		/* this is the key to the whole thing */
+#endif
+
 #include <sys/wait.h>
 #include <sys/time.h>
+#ifndef SYSV
 #include <sys/resource.h>
+#endif
 #include <sys/file.h>
 #include <signal.h>
 #ifdef TERMIO
@@ -49,9 +67,22 @@
 #include <sgtty.h>
 #endif
 #endif
+
+#ifdef SYSV
+#undef TIOCGWINSZ
+#endif
+
 #include <sys/param.h>
 
-#define VERSIONSTR "zsh v2.00.03"
+#ifdef SYSV
+#undef _POSIX_SOURCE
+#endif
+
+#ifdef __hp9000s800
+#include <sys/bsdtty.h>
+#endif
+
+#define VERSIONSTR "zsh v2.1.0"
 
 #if 0 /* __STDC__ */
 #include <unistd.h>
@@ -83,6 +114,13 @@
 #define EXTERN extern
 #endif
 
+#ifndef F_OK
+#define F_OK 00
+#define R_OK 04
+#define W_OK 02
+#define X_OK 01
+#endif
+
 #include "zle.h"
 
 /* size of job list */
@@ -132,6 +170,109 @@ vptr (*ncalloc)DCLPROTO((int));
 #define ALPOPS " \201"
 #define HISTMARK "\201"
 
+#define SEPER 1
+#define NEWLIN 2
+#define LEXERR 3
+#define SEMI 4
+#define DSEMI 5
+#define AMPER 6
+#define INPAR 7
+#define INBRACE 8
+#define OUTPAR 9
+#define DBAR 10
+#define DAMPER 11
+#define BANG 12
+#define OUTBRACE 13
+#define OUTANG 14
+#define OUTANGBANG 15
+#define DOUTANG 16
+#define DOUTANGBANG 17
+#define INANG 18
+#define DINANG 19
+#define DINANGDASH 20
+#define INANGAMP 21
+#define OUTANGAMP 22
+#define OUTANGAMPBANG 23
+#define DOUTANGAMP 24
+#define DOUTANGAMPBANG 25
+#define TRINANG 26
+#define BAR 27
+#define BARAMP 28
+#define DINBRACK 29
+#define DOUTBRACK 30
+#define STRING 31
+#define ENVSTRING 32
+#define ENVARRAY 33
+#define ENDINPUT 34
+#define INOUTPAR 35
+#define DO 36
+#define DONE 37
+#define ESAC 38
+#define THEN 39
+#define ELIF 40
+#define ELSE 41
+#define FI 42
+#define FOR 43
+#define CASE 44
+#define IF 45
+#define WHILE 46
+#define FUNC 47
+#define REPEAT 48
+#define TIME 49
+#define UNTIL 50
+#define EXEC 51
+#define COMMAND 52
+#define SELECT 53
+#define COPROC 54
+#define NOGLOB 55
+#define DASH 56
+#define NOCORRECT 57
+#define FOREACH 58
+#define ZEND 59
+
+#define WRITE 0
+#define WRITENOW 1
+#define APP 2
+#define APPNOW 3
+#define MERGEOUT 4
+#define MERGEOUTNOW 5
+#define ERRAPP 6
+#define ERRAPPNOW 7
+#define READ 8
+#define HEREDOC 9
+#define MERGE 10
+#define CLOSE 11
+#define INPIPE 12
+#define OUTPIPE 13
+#define HERESTR 14
+#define NONE 15
+#define HEREDOCDASH 16
+
+#ifdef GLOBALS
+int redirtab[TRINANG-OUTANG+1] = {
+	WRITE,
+	WRITENOW,
+	APP,
+	APPNOW,
+	READ,
+	HEREDOC,
+	HEREDOCDASH,
+	MERGE,
+	MERGEOUT,
+	MERGEOUTNOW,
+	ERRAPP,
+	ERRAPPNOW,
+	HERESTR,
+};
+#else
+int redirtab[TRINANG-OUTANG+1];
+#endif
+
+#define IS_READFD(X) (((X)>=READ && (X)<=MERGE)||(X)==HEREDOCDASH||(X)==HERESTR)
+#define IS_REDIROP(X) ((X)>=OUTANG && (X)<=TRINANG)
+#define IS_ERROR_REDIR(X) ((X)>=MERGEOUT && (X)<=ERRAPPNOW)
+#define UN_ERROR_REDIR(X) ((X)-MERGEOUT+WRITE)
+
 EXTERN char **environ;
 
 typedef struct hashtab *Hashtab;
@@ -304,27 +445,8 @@ struct cmd {
 struct redir {
 	int type,fd1,fd2,ifil;
 	char *name;
-	struct redir *pair;		/* used by the parser */
    };
 
-#define WRITE 0
-#define WRITENOW 1
-#define APP 2
-#define APPNOW 3
-#define MERGEOUT 4
-#define MERGEOUTNOW 5
-#define ERRAPP 6
-#define ERRAPPNOW 7
-#define READ 8
-#define HEREDOC 9
-#define MERGE 10
-#define CLOSE 11
-#define INPIPE 12
-#define OUTPIPE 13
-#define HERESTR 14
-#define NONE 15
-#define HEREDOCDASH 16
-
 /* tree element for conditionals */
 
 struct cond {
@@ -515,7 +637,6 @@ struct job {
 											this job was spawned */
 	struct process *procs;		/* list of processes */
 	Lklist filelist;				/* list of files to delete when done */
-	struct ttyinfo ttyinfo;		/* saved tty state */
 	};
 
 #define STAT_CHANGED 1		/* status changed and not reported */
@@ -665,69 +786,91 @@ struct option {
 #define GLOBCOMPLETE '\6'
 #define CSHJUNKIEQUOTES '\7'
 #define PUSHDMINUS '\10'
+#define CSHJUNKIELOOPS '\11'
+#define RCQUOTES '\12'
+#define KSHOPTIONPRINT '\13'
+#define NOSHORTLOOPS '\14'
+#define LASTMENU '\15'
+#define AUTOMENU '\16'
+#define HISTVERIFY '\17'
+#define NOLISTBEEP '\20'
+#define NOHUP '\21'
+#define NOEQUALS '\22'
+#define CSHNULLGLOB '\23'
 
 #ifndef GLOBALS
 extern struct option optns[];
 #else
 struct option optns[] = {
-	"correct",'0',
-	"noclobber",'1',
-	"nobadpattern",'2',
-	"nonomatch",'3',
-	"globdots",'4',
-	"notify",'5',
-	"bgnice",'6',
-	"ignoreeof",'7',
-	"markdirs",'8',
-	"autolist",'9',
-	"nobeep",'B',
-	"printexitvalue",'C',
-	"pushdtohome",'D',
-	"pushdsilent",'E',
-	"noglob",'F',
-	"nullglob",'G',
-	"rmstarsilent",'H',
-	"ignorebraces",'I',
-	"autocd",'J',
-	"nobanghist",'K',
-	"sunkeyboardhack",'L',
-	"singlelinezle",'M',
-	"autopushd",'N',
-	"correctall",'O',
-	"rcexpandparam",'P',
-	"pathdirs",'Q',
-	"longlistjobs",'R',
-	"recexact",'S',
-	"cdablevars",'T',
-	"mailwarning",'U',
-	"nopromptclobber",'V',
-	"autoresume",'W',
-	"listtypes",'X',
-	"menucomplete",'Y',
-	"zle",'Z',
-	"allexport",'a',
-	"errexit",'e',
-	"norcs",'f',
-	"histignorespace",'g',
-	"histignoredups",'h',
-	"interactive",'i',
-	"histlit",'j',
-	"interactivecomments",'k',
-	"login",'l',
-	"monitor",'m',
-	"noexec",'n',
-	"shinstdin",'s',
-	"nounset",'u',
-	"verbose",'v',
-	"chaselinks",'w',
-	"xtrace",'x',
-	"shwordsplit",'y',
-	"menucompletebeep",'\2',
-	"histnostore",'\3',
-	"extendedglob",'\5',
-	"globcomplete",'\6',
-	"cshjunkiequotes",'\7',
-	"pushdminus",'\10',
+	"correct",CORRECT,
+	"noclobber",NOCLOBBER,
+	"nobadpattern",NOBADPATTERN,
+	"nonomatch",NONOMATCH,
+	"globdots",GLOBDOTS,
+	"notify",NOTIFY,
+	"bgnice",BGNICE,
+	"ignoreeof",IGNOREEOF,
+	"markdirs",MARKDIRS,
+	"autolist",AUTOLIST,
+	"nobeep",NOBEEP,
+	"printexitvalue",PRINTEXITVALUE,
+	"pushdtohome",PUSHDTOHOME,
+	"pushdsilent",PUSHDSILENT,
+	"noglob",NOGLOBOPT,
+	"nullglob",NULLGLOB,
+	"rmstarsilent",RMSTARSILENT,
+	"ignorebraces",IGNOREBRACES,
+	"autocd",AUTOCD,
+	"nobanghist",NOBANGHIST,
+	"sunkeyboardhack",SUNKEYBOARDHACK,
+	"singlelinezle",SINGLELINEZLE,
+	"autopushd",AUTOPUSHD,
+	"correctall",CORRECTALL,
+	"rcexpandparam",RCEXPANDPARAM,
+	"pathdirs",PATHDIRS,
+	"longlistjobs",LONGLISTJOBS,
+	"recexact",RECEXACT,
+	"cdablevars",CDABLEVARS,
+	"mailwarning",MAILWARNING,
+	"nopromptclobber",NOPROMPTCLOBBER,
+	"autoresume",AUTORESUME,
+	"listtypes",LISTTYPES,
+	"menucomplete",MENUCOMPLETE,
+	"zle",USEZLE,
+	"allexport",ALLEXPORT,
+	"errexit",ERREXIT,
+	"norcs",NORCS,
+	"histignorespace",HISTIGNORESPACE,
+	"histignoredups",HISTIGNOREDUPS,
+	"interactive",INTERACTIVE,
+	"histlit",HISTLIT,
+	"interactivecomments",INTERACTIVECOMMENTS,
+	"login",LOGINSHELL,
+	"monitor",MONITOR,
+	"noexec",NOEXEC,
+	"shinstdin",SHINSTDIN,
+	"nounset",NOUNSET,
+	"verbose",VERBOSE,
+	"chaselinks",CHASELINKS,
+	"xtrace",XTRACE,
+	"shwordsplit",SHWORDSPLIT,
+	"menucompletebeep",MENUCOMPLETEBEEP,
+	"histnostore",HISTNOSTORE,
+	"extendedglob",EXTENDEDGLOB,
+	"globcomplete",GLOBCOMPLETE,
+	"cshjunkiequotes",CSHJUNKIEQUOTES,
+	"pushdminus",PUSHDMINUS,
+	"cshjunkieloops",CSHJUNKIELOOPS,
+	"rcquotes",RCQUOTES,
+	"noshortloops",NOSHORTLOOPS,
+	"lastmenu",LASTMENU,
+	"automenu",AUTOMENU,
+	"histverify",HISTVERIFY,
+	"nolistbeep",NOLISTBEEP,
+	"nohup",NOHUP,
+	"noequals",NOEQUALS,
+	"kshoptionprint",KSHOPTIONPRINT,
+	"cshnullglob",CSHNULLGLOB,
 	NULL,0
 };
 #endif
@@ -795,6 +938,7 @@ EXTERN FILE *bshin;
 /* NULL-terminated arrays containing path, cdpath, etc. */
 
 EXTERN char **path,**cdpath,**fpath,**watch,**mailpath,**tildedirs,**fignore;
+EXTERN char **hosts,**hostcmds,**optcmds,**bindcmds,**varcmds;
 
 /* named directories */
 
@@ -812,6 +956,9 @@ EXTERN char *yytext;
 
 EXTERN int errflag;
 
+EXTERN char *tokstr;
+EXTERN int tok, tokfd;
+
 /* lexical analyzer error flag */
 
 EXTERN int lexstop;
@@ -850,14 +997,9 @@ EXTERN int histdone;
 
 EXTERN int defev;
 
-/* != 0 if we are in the middle of parsing a command (that is,
-	we have already read the command word */
+/* != 0 if we are about to read a command word */
 
-EXTERN int incmd;
-
-/* != 0 if we just read a redirection */
-
-EXTERN int inredir;
+EXTERN int incmdpos;
 
 /* != 0 if we are in the middle of a [[ ... ]] */
 
@@ -865,16 +1007,16 @@ EXTERN int incond;
 
 /* != 0 if we are about to read a case pattern */
 
-EXTERN int incase;
-
-/* != 0 if we just read a separator (; or newline) */
-
-EXTERN int lsep;
+EXTERN int incasepat;
 
 /* != 0 if we just read FUNCTION */
 
 EXTERN int infunc;
 
+/* != 0 if we just read a newline */
+
+EXTERN int isnewlin;
+
 /* the lists of history events */
 
 EXTERN Lklist histlist,lithistlist;
@@ -1022,13 +1164,17 @@ EXTERN char *rprompt;	/* $RPROMPT */
 EXTERN char *prompt2;	/* etc. */
 EXTERN char *prompt3;
 EXTERN char *prompt4;
+EXTERN char *sprompt;
 EXTERN char *timefmt;
 EXTERN char *watchfmt;
 EXTERN char *wordchars;
 EXTERN char *fceditparam;
+EXTERN char *tmpprefix;
 
 EXTERN char *argzero;	/* $0 */
 
+EXTERN char *hackzero;
+
 /* the hostname */
 
 EXTERN char *hostnam;
@@ -1107,7 +1253,9 @@ EXTERN int lastbase;		/* last input base we used */
 
 /* the limits for child processes */
 
+#ifdef RLIM_INFINITY
 EXTERN struct rlimit limits[RLIM_NLIMITS];
+#endif
 
 /* the current word in the history List */
 
@@ -1133,6 +1281,10 @@ EXTERN char *term;
 
 EXTERN int termok;
 
+/* flag for CSHNULLGLOB */
+
+EXTERN int badcshglob;
+
 /* max size of hline */
 
 EXTERN int hlinesz;
@@ -1201,6 +1353,8 @@ EXTERN int zleparse;
 
 EXTERN int wordbeg;
 
+EXTERN int parbegin;
+
 /* interesting termcap strings */
 
 #define TCCLEARSCREEN 0
@@ -1243,6 +1397,30 @@ extern char *tccapnams[TC_COUNT];
 
 #define tccan(X) (!!tclen[X])
 
+#define HISTFLAG_DONE   1
+#define HISTFLAG_NOEXEC 2
+
 #include "ztype.h"
 #include "funcs.h"
 
+#ifdef __hpux
+#define setpgrp setpgid
+#endif
+
+#define _INCLUDE_POSIX_SOURCE
+#define _INCLUDE_XOPEN_SOURCE
+#define _INCLUDE_HPUX_SOURCE
+
+#ifdef SV_BSDSIG
+#define SV_INTERRUPT SV_BSDSIG
+#endif
+
+#ifdef SYSV
+#define blockchld() sighold(SIGCHLD)
+#define unblockchld() sigrelse(SIGCHLD)
+#define chldsuspend() sigpause(SIGCHLD)
+#else
+#define blockchld() sigblock(sigmask(SIGCHLD))
+#define unblockchld() sigsetmask(0)
+#define chldsuspend() sigpause(0)
+#endif
